import logging

LOG = logging.getLogger(__name__)


def format_info(variant, variant_type="snv", nr_cases=None, add_freq=False):
    """Format the info field for SNV variants

    Args:
        variant(dict)
        variant_type(str): snv or sv
        nr_cases(int)

    Returns:
        vcf_info(str): A VCF formated info field
    """

    observations = variant.get("observations", 0)

    homozygotes = variant.get("homozygote")
    hemizygotes = variant.get("hemizygote")

    vcf_info = f"Obs={observations}"
    if homozygotes:
        vcf_info += f";Hom={homozygotes}"
    if hemizygotes:
        vcf_info += f";Hem={hemizygotes}"

    if add_freq and nr_cases and nr_cases > 0:
        frequency = observations / nr_cases
        vcf_info += f";Frq={frequency:.5f}"

    # This is SV specific
    if variant_type == "sv":
        pos = int((variant["pos_left"] + variant["pos_right"]) / 2)
        end = int((variant["end_left"] + variant["end_right"]) / 2)

        if not variant["sv_type"] == "BND":
            vcf_info += f";SVTYPE={variant['sv_type']};END={end};SVLEN={variant['length']}"
        else:
            vcf_info += ";SVTYPE=BND"

    return vcf_info


def format_variant(variant, variant_type="snv", nr_cases=None, add_freq=False):
    """Convert variant information to a VCF formated string

    Args:
        variant(dict)
        variant_type(str)
        nr_cases(int)

    Returns:
        vcf_variant(str)
    """
    chrom = variant.get("chrom")
    pos = variant.get("start")

    ref = variant.get("ref")
    alt = variant.get("alt")

    if variant_type == "sv":
        pos = int((variant["pos_left"] + variant["pos_right"]) / 2)
        end = int((variant["end_left"] + variant["end_right"]) / 2)
        ref = "N"

        if not variant["sv_type"] == "BND":
            alt = f"<{variant['sv_type']}>"
        else:
            alt = f"N]{variant['end_chrom']}:{end}]"

    info = None

    info = format_info(variant, variant_type=variant_type, nr_cases=nr_cases, add_freq=add_freq)

    return f"{chrom}\t{pos}\t.\t{ref}\t{alt}\t.\t.\t{info}"



import copy

from loqusdb.build_models.variant import build_variant
from loqusdb.constants import GRCH37, GRCH38


class TestInsertVariant:
    def test_insert_one_variant(self, mongo_adapter, simplest_variant):
        """Test to insert one variant"""

        mongo_adapter.add_variant(simplest_variant)

        db = mongo_adapter.db

        # Get the variant without the adapter
        mongo_variant = db.variant.find_one()

        assert mongo_variant["_id"] == "test"
        assert mongo_variant["observations"] == 1
        assert mongo_variant["homozygote"] == 0

    def test_insert_one_variant_twice(self, mongo_adapter, simplest_variant):
        """Test to insert one variant"""

        mongo_adapter.add_variant(simplest_variant)
        mongo_adapter.add_variant(simplest_variant)

        db = mongo_adapter.db

        mongo_variant = db.variant.find_one()

        assert mongo_variant["_id"] == "test"
        assert mongo_variant["observations"] == 2
        assert mongo_variant.get("homozygote", 0) == 0

    def test_insert_hom_variant(self, real_mongo_adapter, homozygous_variant):
        """Test to insert a homozygote variant"""
        mongo_adapter = real_mongo_adapter
        mongo_adapter.add_variant(homozygous_variant)

        db = mongo_adapter.db

        mongo_variant = db.variant.find_one()
        assert mongo_variant["_id"] == "test"
        assert mongo_variant["observations"] == 1
        assert mongo_variant.get("homozygote", 0) == 1
        assert mongo_variant["families"] == ["1"]

    def test_insert_many(self, mongo_adapter, simplest_variant):
        """Test to insert a homozygote variant"""

        for _ in range(10000):
            mongo_adapter.add_variant(simplest_variant)

        db = mongo_adapter.db

        mongo_variant = db.variant.find_one()
        assert mongo_variant["_id"] == "test"
        assert mongo_variant["observations"] == 10000
        assert mongo_variant.get("homozygote", 0) == 0


class TestGetVariant:
    def test_get_variant(self, mongo_client, mongo_adapter, simplest_variant):
        """Test to insert one variant"""

        # Insert without adapter
        db = mongo_client["test"]
        db.variant.insert_one(simplest_variant)

        mongo_variant = mongo_adapter.get_variant(simplest_variant)
        assert mongo_variant["_id"] == "test"

    def test_get_none(self, mongo_adapter, simplest_variant):
        """Test to get non existing variant"""

        mongo_variant = mongo_adapter.get_variant(simplest_variant)

        assert mongo_variant is None


class TestBulkOperations:
    def test_insert_one_variant(self, real_mongo_adapter, simplest_variant):
        """Test to insert one variant with bulk insert"""

        mongo_adapter = real_mongo_adapter
        variants = [simplest_variant]

        mongo_adapter.add_variants(variants)
        db = mongo_adapter.db
        mongo_variant = db.variant.find_one()

        assert mongo_variant["_id"] == "test"
        assert mongo_variant["observations"] == 1
        assert mongo_variant["homozygote"] == 0

    def test_insert_two_variants(self, real_mongo_adapter):
        """Test to insert two variants with bulk"""

        adapter = real_mongo_adapter
        db = adapter.db

        variants = [
            {"_id": "test", "homozygote": 0},
            {"_id": "test_1", "homozygote": 1},
        ]

        adapter.add_variants(variants)

        first_variant = db.variant.find_one({"_id": "test"})
        second_variant = db.variant.find_one({"_id": "test_1"})

        assert first_variant["_id"] == "test"
        assert first_variant["observations"] == 1
        assert first_variant.get("homozygote", 0) == 0

        assert second_variant["_id"] == "test_1"
        assert second_variant["observations"] == 1
        assert second_variant.get("homozygote", 0) == 1

    def test_insert_many(self, real_mongo_adapter):
        adapter = real_mongo_adapter
        db = adapter.db

        variants = ({"_id": "test", "homozygote": 0} for i in range(20000))

        adapter.add_variants(variants)

        mongo_variant = db.variant.find_one()

        assert mongo_variant["_id"] == "test"
        assert mongo_variant["observations"] == 20000
        assert mongo_variant.get("homozygote", 0) == 0


class TestRemoveVariant:
    def test_remove_one_variant(self, mongo_adapter):
        """Test to update one variant"""

        db = mongo_adapter.db

        variant = {"_id": "test", "observations": 1}

        db.variant.insert_one(variant)

        mongo_adapter.delete_variant(variant)

        assert db.variant.find_one() is None

    def test_downcount_one_variant(self, mongo_adapter):
        """Test to update one variant"""

        db = mongo_adapter.db

        insert_variant = {"_id": "test", "families": ["1", "2"], "observations": 2}

        db.variant.insert_one(insert_variant)

        variant = {"_id": "test", "case_id": "1"}

        mongo_adapter.delete_variant(variant)

        mongo_variant = db.variant.find_one()

        assert mongo_variant["observations"] == 1
        assert mongo_variant["families"] == ["2"]

    def test_remove_non_existing(self, mongo_adapter, simplest_variant):
        db = mongo_adapter.db

        mongo_adapter.delete_variant(simplest_variant)

        assert db.variant.find_one() is None


class TestRemoveSV:
    def test_remove_one_SV(self, mongo_adapter, del_variant, case_obj):
        # GIVEN a database poulated with one SV
        db = mongo_adapter.db
        formated_variant = build_variant(
            del_variant, case_obj=case_obj, case_id=case_obj["case_id"], genome_build=GRCH37
        )
        mongo_adapter.add_structural_variant(formated_variant)
        mongo_SV = db.structural_variant.find_one()
        mongo_identity = db.identity.find_one()
        assert mongo_SV is not None
        assert mongo_identity is not None
        # WHEN deleting SV
        mongo_adapter.delete_structural_variant(formated_variant)

        # THEN there should be no remaining SVs in the database
        mongo_SV = db.structural_variant.find_one()
        mongo_identity = db.indentity.find_one()
        assert mongo_SV is None
        assert mongo_identity is None

    def test_remove_one_of_two_SV(self, mongo_adapter, duptandem_variant, case_obj):
        # GIVEN a database poulated with one SV
        db = mongo_adapter.db
        formated_variant = build_variant(
            duptandem_variant, case_obj=case_obj, case_id=case_obj["case_id"], genome_build=GRCH37
        )
        mongo_adapter.add_structural_variant(formated_variant)

        # Add second of same variant, changing the start and end position slightly
        formated_variant_ = copy.deepcopy(formated_variant)
        formated_variant_["pos"] = formated_variant_["pos"] + 2
        formated_variant_["end"] = formated_variant_["end"] - 1
        formated_variant_["case_id"] = "case_2"
        mongo_adapter.add_structural_variant(formated_variant_)

        # This should correspond to one structural variant document
        mongo_svs = list(db.structural_variant.find())
        assert len(mongo_svs) == 1
        mongo_sv = mongo_svs[0]
        assert mongo_sv["pos_sum"] == formated_variant["pos"] + formated_variant_["pos"]
        # And two identity documents
        mongo_identities = list(db.identity.find())
        assert len(mongo_identities) == 2

        # WHEN deleting the variant from the first case
        mongo_adapter.delete_structural_variant(formated_variant)

        # THEN the SV document should have the pos_sum equal to the pos of the
        # SV from the second case
        mongo_svs = list(db.structural_variant.find())
        assert len(mongo_svs) == 1
        mongo_sv = mongo_svs[0]
        assert mongo_sv["pos_sum"] == formated_variant_["pos"]
        # And one identity documents
        mongo_identities = list(db.identity.find())
        assert len(mongo_identities) == 1


class TestHelperMethods:
    def test_update_sv_metrics(self, mongo_adapter):
        # GIVEN a mongo adapter

        # WHEN cluster_len > 10000
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(
            sv_type="INV", pos_mean=10000, end_mean=30000, max_window=3000
        )
        # THEN interval_size should be the cluster_len divided by 10
        assert cluster_len == 20000
        assert interval_size == round(20000 / 10, -2)

        # WHEN cluster_len <10000
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(
            sv_type="DUP", pos_mean=10000, end_mean=15000, max_window=3000
        )

        # THEN interval_size should be cluster_len divided by 5
        assert cluster_len == 5000
        assert interval_size == round(5000 / 5, -2)

        # WHEN interval_size < 1000
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(
            sv_type="DEL", pos_mean=10000, end_mean=10500, max_window=3000
        )

        # THEN interval_size should be cluster_len divided by 2
        assert cluster_len == 500
        assert interval_size == round(500 / 2, -2)

        # WHEN interval size is > max_window
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(
            sv_type="INV", pos_mean=100000, end_mean=200000, max_window=3000
        )
        # THEN interval size should be set to max_window
        assert cluster_len == 100000
        assert interval_size == 3000

        # WHEN sv_type == BND
        cluster_len, interval_size = mongo_adapter._update_sv_metrics(
            sv_type="BND", pos_mean=1000, end_mean=2000, max_window=3000
        )
        # THEN cluster_len should be 10e10 and interval_size 2*max window
        assert cluster_len == 10e10
        assert interval_size == 2 * 3000


class TestRemoveSV_grch38:
    def test_remove_one_SV(self, mongo_adapter, chr_del_variant, case_obj):
        # GIVEN a database poulated with one SV
        db = mongo_adapter.db
        formated_variant = build_variant(
            chr_del_variant, case_obj=case_obj, case_id=case_obj["case_id"], genome_build=GRCH38
        )
        mongo_adapter.add_structural_variant(formated_variant)
        mongo_SV = db.structural_variant.find_one()
        mongo_identity = db.identity.find_one()
        assert mongo_SV is not None
        assert mongo_identity is not None
        # WHEN deleting SV
        mongo_adapter.delete_structural_variant(formated_variant)

        # THEN there should be no remaining SVs in the database
        mongo_SV = db.structural_variant.find_one()
        mongo_identity = db.indentity.find_one()
        assert mongo_SV is None
        assert mongo_identity is None

    def test_remove_one_of_two_SV(self, mongo_adapter, chr_duptandem_variant, case_obj):
        # GIVEN a database poulated with one SV
        db = mongo_adapter.db
        formated_variant = build_variant(
            chr_duptandem_variant,
            case_obj=case_obj,
            case_id=case_obj["case_id"],
            genome_build=GRCH38,
        )
        mongo_adapter.add_structural_variant(formated_variant)

        # Add second of same variant, changing the start and end position slightly
        formated_variant_ = copy.deepcopy(formated_variant)
        formated_variant_["pos"] = formated_variant_["pos"] + 2
        formated_variant_["end"] = formated_variant_["end"] - 1
        formated_variant_["case_id"] = "case_2"
        mongo_adapter.add_structural_variant(formated_variant_)

        # This should correspond to one structural variant document
        mongo_svs = list(db.structural_variant.find())
        assert len(mongo_svs) == 1
        mongo_sv = mongo_svs[0]
        assert mongo_sv["pos_sum"] == formated_variant["pos"] + formated_variant_["pos"]
        # And two identity documents
        mongo_identities = list(db.identity.find())
        assert len(mongo_identities) == 2

        # WHEN deleting the variant from the first case
        mongo_adapter.delete_structural_variant(formated_variant)

        # THEN the SV document should have the pos_sum equal to the pos of the
        # SV from the second case
        mongo_svs = list(db.structural_variant.find())
        assert len(mongo_svs) == 1
        mongo_sv = mongo_svs[0]
        assert mongo_sv["pos_sum"] == formated_variant_["pos"]
        # And one identity documents
        mongo_identities = list(db.identity.find())
        assert len(mongo_identities) == 1



import logging

from loqusdb import INDEXES
from mongo_adapter import MongoAdapter as BaseAdapter

from loqusdb.plugins.mongo.case import CaseMixin
from loqusdb.plugins.mongo.profile_variant import ProfileVariantMixin
from loqusdb.plugins.mongo.variant import VariantMixin

logger = logging.getLogger(__name__)


class MongoAdapter(BaseAdapter, VariantMixin, CaseMixin, ProfileVariantMixin):
    """docstring for MongoAdapter"""

    def wipe_db(self):
        """Wipe the whole database"""
        logger.warning("Wiping the whole database")
        self.client.drop_database(self.db_name)
        logger.debug("Database wiped")

    def indexes(self, collection=None):
        """Return a list with the current indexes

        Skip the mandatory _id_ indexes

        Args:
            collection(str)

        Returns:
            indexes(list)
        """
        indexes = []
        for collection_name in self.db.list_collection_names():
            if collection and collection != collection_name:
                continue
            for index_name in self.db[collection_name].index_information():
                if index_name != "_id_":
                    indexes.append(index_name)
        return indexes

    def check_indexes(self):
        """Check if the indexes exists"""
        for collection_name in INDEXES:
            existing_indexes = self.indexes(collection_name)
            indexes = INDEXES[collection_name]
            for index in indexes:
                index_name = index.document.get("name")
                if index_name not in existing_indexes:
                    logger.warning(
                        "Index {0} missing. Run command `loqusdb index`".format(index_name)
                    )
                    return
        logger.info("All indexes exists")

    def ensure_indexes(self):
        """Update the indexes"""
        for collection_name in INDEXES:
            existing_indexes = self.indexes(collection_name)
            indexes = INDEXES[collection_name]
            for index in indexes:
                index_name = index.document.get("name")
                if index_name in existing_indexes:
                    logger.debug("Index exists: %s" % index_name)
                    self.db[collection_name].drop_index(index_name)
            logger.info(
                "creating indexes for collection {0}: {1}".format(
                    collection_name,
                    ", ".join(index.document.get("name") for index in indexes),
                )
            )

            self.db[collection_name].create_indexes(indexes)



import pytest
from loqusdb.exceptions import CaseError
from loqusdb.utils.case import update_case


def test_update_case(case_obj, sv_case_obj):
    ## GIVEN a case with snv info and a case with sv info
    assert sv_case_obj["vcf_path"] is None
    assert case_obj["vcf_path"] is not None
    assert case_obj["vcf_sv_path"] is None
    assert sv_case_obj["vcf_sv_path"] is not None

    ## WHEN updating the case
    updated_case = update_case(case_obj=sv_case_obj, existing_case=case_obj)

    ## THEN assert that the cases have been merged without affecting the original cases
    assert updated_case["vcf_path"] is not None
    assert updated_case["vcf_path"] == case_obj["vcf_path"]
    assert sv_case_obj["vcf_path"] is None

    assert updated_case["vcf_sv_path"] is not None
    assert updated_case["vcf_sv_path"] == sv_case_obj["vcf_sv_path"]
    assert case_obj["vcf_sv_path"] is None

    assert updated_case["nr_variants"] == case_obj["nr_variants"]
    assert updated_case["nr_sv_variants"] == sv_case_obj["nr_sv_variants"]


def test_update_existing_vcf(case_obj):
    ## GIVEN a case object with VCF information
    ## WHEN trying to update existing VCF information
    ## THEN assert that a CaseError is raised
    with pytest.raises(CaseError):
        updated_case = update_case(case_obj, case_obj)



import pytest
from loqusdb.exceptions import CaseError
from loqusdb.utils.load import load_database
from loqusdb.constants import GRCH37, GRCH38


def test_load_database(vcf_path, ped_path, real_mongo_adapter, case_id):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type="ped",
        genome_build=GRCH37,
    )

    mongo_case = db.case.find_one()

    assert mongo_case["case_id"] == case_id


def test_load_database_alternative_ped(vcf_path, ped_path, real_mongo_adapter, case_id):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type="ped",
        case_id="alternative",
        genome_build=GRCH37,
    )

    mongo_case = db.case.find_one()
    mongo_variant = db.variant.find_one()

    assert mongo_case["case_id"] == "alternative"
    assert mongo_variant["families"] == ["alternative"]


def test_load_database_wrong_ped(vcf_path, funny_ped_path, real_mongo_adapter):
    mongo_adapter = real_mongo_adapter
    ## GIVEN a vcf and ped file with wrong individuals
    ## WHEN loading the information
    ## THEN Error should be raised since individuals is not in vcf
    with pytest.raises(CaseError):
        load_database(
            adapter=mongo_adapter,
            variant_file=vcf_path,
            family_file=funny_ped_path,
            family_type="ped",
            genome_build=GRCH37,
        )


def test_load_database_grch38(vcf_path, ped_path, real_mongo_adapter, case_id):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type="ped",
        genome_build=GRCH38,
    )

    mongo_case = db.case.find_one()

    assert mongo_case["case_id"] == case_id


def test_load_database_alternative_ped_grch38(vcf_path, ped_path, real_mongo_adapter, case_id):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type="ped",
        case_id="alternative",
        genome_build=GRCH38,
    )

    mongo_case = db.case.find_one()
    mongo_variant = db.variant.find_one()

    assert mongo_case["case_id"] == "alternative"
    assert mongo_variant["families"] == ["alternative"]


def test_load_database_wrong_ped_grch38(vcf_path, funny_ped_path, real_mongo_adapter):
    mongo_adapter = real_mongo_adapter
    ## GIVEN a vcf and ped file with wrong individuals
    ## WHEN loading the information
    ## THEN Error should be raised since individuals is not in vcf
    with pytest.raises(CaseError):
        load_database(
            adapter=mongo_adapter,
            variant_file=vcf_path,
            family_file=funny_ped_path,
            family_type="ped",
            genome_build=GRCH38,
        )



import logging

import click

from loqusdb.commands.cli import cli as base_command

LOG = logging.getLogger(__name__)


@base_command.command("identity", short_help="Search identity collection")
@click.option("-v", "--variant-id", help="Search for a variant ID.")
@click.pass_context
def identity(ctx, variant_id):
    """Check how well SVs are working in the database"""
    if not variant_id:
        LOG.warning("Please provide a variant id")
        ctx.abort()

    adapter = ctx.obj["adapter"]

    LOG.info("Search variants {0}".format(adapter))

    variant_count: int = adapter.db.identity.count_documents({"variant_id": variant_id})
    if variant_count == 0:
        LOG.info(f"No hits for variant {variant_id}")
        return

    for res in result:
        click.echo(res)



from click.testing import CliRunner

from loqusdb.commands.cli import cli as base_command
from loqusdb.plugins.mongo.adapter import MongoAdapter


def test_view_cases_base(real_mongo_adapter: MongoAdapter, real_db_name: str):
    """Test the base command that returns database cases."""

    ## GIVEN an empty database
    assert sum([1 for _ in real_mongo_adapter.cases()]) == 0

    runner = CliRunner()

    # THEN the case command should return No cases found error
    command = ["--database", real_db_name, "cases"]
    result = runner.invoke(base_command, command)
    assert result.exit_code == 1
    assert "No cases found in database" in result.output



import pytest

from loqusdb.exceptions import CaseError


class TestDeleteCase:
    def test_delete_one_case(self, mongo_adapter, case_obj):
        """Test to delete one case"""

        db = mongo_adapter.db

        mongo_case = db.case.find_one()

        assert mongo_case == None

        db.case.insert_one(case_obj)

        mongo_case = db.case.find_one()

        assert mongo_case["case_id"] == case_obj["case_id"]
        assert mongo_case["vcf_path"] == case_obj["vcf_path"]

        mongo_adapter.delete_case(case_obj)

        assert db.case.find_one() == None

    def test_delete_non_existing(self, mongo_adapter, case_obj):
        """Test to delete non exsting case"""

        with pytest.raises(CaseError):
            mongo_adapter.delete_case(case_obj)


class TestInsertCase:
    def test_insert_one_case(self, mongo_adapter, case_obj):
        db = mongo_adapter.db

        mongo_adapter.add_case(case_obj)

        mongo_case = db.case.find_one()

        assert mongo_case["case_id"] == case_obj["case_id"]
        assert mongo_case["vcf_path"] == case_obj["vcf_path"]

    def test_insert_one_case_twice(self, mongo_adapter, case_obj):
        db = mongo_adapter.db

        mongo_adapter.add_case(case_obj)

        with pytest.raises(CaseError):
            mongo_adapter.add_case(case_obj)


class TestGetCase:
    def test_get_case(self, mongo_adapter, case_obj):
        """Test to get non existing case"""

        db = mongo_adapter.db
        db.case.insert_one(case_obj)

        mongo_case = mongo_adapter.case(case_obj)

        assert mongo_case["case_id"] == case_obj["case_id"]
        assert mongo_case["vcf_path"] == case_obj["vcf_path"]

    def test_get_non_existing(self, mongo_adapter, simple_case):
        """Test to get non existing case"""

        db = mongo_adapter.db

        mongo_case = mongo_adapter.case(simple_case)

        assert mongo_case == None

    def test_get_multiple_cases(self, mongo_adapter):
        """Test to get non existing case"""

        db = mongo_adapter.db
        case_1 = {"case_id": "test", "vcf_path": "test.vcf"}

        case_2 = {"case_id": "test2", "vcf_path": "test2.vcf"}

        db.case.insert_one(case_1)
        db.case.insert_one(case_2)

        mongo_cases = mongo_adapter.cases()

        assert sum(1 for i in mongo_cases) == 2

    def test_nr_cases(self, mongo_adapter):
        """Test to get non existing case"""

        db = mongo_adapter.db
        case_1 = {"case_id": "test", "vcf_path": "test.vcf", "vcf_sv_path": None}

        db.case.insert_one(case_1)

        assert mongo_adapter.nr_cases() == 1
        assert mongo_adapter.nr_cases(sv_cases=True) == 0
        assert mongo_adapter.nr_cases(snv_cases=True) == 1
        assert mongo_adapter.nr_cases(snv_cases=True, sv_cases=True) == 1

    def test_nr_cases_only_snv(self, mongo_adapter):
        """Test to get non existing case"""
        db = mongo_adapter.db
        case_1 = {"case_id": "test", "vcf_path": "test.vcf"}

        db.case.insert_one(case_1)

        assert mongo_adapter.nr_cases() == 1
        assert mongo_adapter.nr_cases(sv_cases=True) == 0
        assert mongo_adapter.nr_cases(snv_cases=True) == 1

    def test_nr_cases_sv(self, mongo_adapter):
        """Test to get non existing case"""

        db = mongo_adapter.db
        case_1 = {"case_id": "test", "vcf_sv_path": "test.vcf"}

        db.case.insert_one(case_1)

        assert mongo_adapter.nr_cases() == 1
        assert mongo_adapter.nr_cases(sv_cases=True) == 1
        assert mongo_adapter.nr_cases(snv_cases=True) == 0

    def test_get_multiple_cases(self, mongo_adapter):
        """Test to get non existing case"""

        db = mongo_adapter.db
        case_1 = {"case_id": "test", "vcf_path": "test.vcf", "vcf_sv_path": "test.vcf"}

        case_2 = {"case_id": "test2", "vcf_path": "test2.vcf"}

        case_3 = {"case_id": "test3", "vcf_path": "test3.vcf"}

        db.case.insert_one(case_1)
        db.case.insert_one(case_2)
        db.case.insert_one(case_3)

        assert mongo_adapter.nr_cases() == 3
        assert mongo_adapter.nr_cases(snv_cases=True) == 3
        assert mongo_adapter.nr_cases(sv_cases=True) == 1
        assert mongo_adapter.nr_cases(sv_cases=True, snv_cases=True) == 3



import pytest
from loqusdb.exceptions import VcfError
from loqusdb.utils.vcf import check_vcf


def test_check_vcf_correct(vcf_path):
    ## GIVEN a vcf file and a counter that checks the number of variants
    true_nr = 0

    with open(vcf_path, "r") as f:
        for line in f:
            if not line.startswith("#"):
                true_nr += 1

    ## WHEN collecting the VCF info
    vcf_info = check_vcf(vcf_path)

    ## THEN assert that the number of variants collected is correct
    assert vcf_info["nr_variants"] == true_nr
    ## THEN assert that the variant type is correct
    assert vcf_info["variant_type"] == "snv"


def test_check_vcf_double_variant(double_vcf_path):
    ## GIVEN a variant file where a variant is duplicated
    ## WHEN checking the vcf
    ## THEN assert that the function raises a VcfError
    with pytest.raises(VcfError):
        check_vcf(double_vcf_path)


def test_check_vcf_unsorted(unsorted_vcf_path):
    ## GIVEN a vcf file with unsorted variants
    ## WHEN checking the vcf
    ## THEN assert that the function raises a VcfError
    with pytest.raises(VcfError):
        check_vcf(unsorted_vcf_path)


def test_check_sv_vcf(sv_vcf_path):
    ## GIVEN a vcf file and a counter that checks the number of variants
    true_nr = 0

    with open(sv_vcf_path, "r") as f:
        for line in f:
            if not line.startswith("#"):
                true_nr += 1

    ## WHEN collecting the VCF info
    vcf_info = check_vcf(sv_vcf_path, expected_type="sv")

    ## THEN assert that the number of variants collected is correct
    assert vcf_info["nr_variants"] == true_nr
    ## THEN assert that the variant type is correct
    assert vcf_info["variant_type"] == "sv"


def test_check_vcf_wrong_type(sv_vcf_path):
    ## GIVEN a sv vcf file

    ## WHEN collecting the VCF info with wrong variant type
    ## THEN assert that a VcfError is raised
    with pytest.raises(VcfError):
        vcf_info = check_vcf(sv_vcf_path, "snv")



import logging

import cyvcf2
import pytest

from loqusdb.build_models.case import build_case
from loqusdb.build_models.variant import build_variant
from loqusdb.log import init_log
from loqusdb.models import Case
from loqusdb.plugins.mongo.adapter import MongoAdapter
from loqusdb.utils.load import update_case
from mongomock import MongoClient as MockClient
from ped_parser import FamilyParser
from pymongo import MongoClient

logger = logging.getLogger(".")

init_log(logger, loglevel="DEBUG")

REAL_DATABASE = "loqus-test"
TEST_DATABASE = "test"


class CyvcfVariant(object):
    """Mock a cyvcf variant

    Default is to return a variant with three individuals high genotype
    quality.
    """

    def __init__(
        self,
        chrom="1",
        pos=80000,
        ref="A",
        alt="C",
        end=None,
        gt_quals=[60, 60, 60],
        gt_types=[1, 1, 0],
        var_type="snv",
        var_id=None,
        info_dict={},
    ):
        super(CyvcfVariant, self).__init__()
        self.CHROM = chrom
        self.POS = pos
        self.REF = ref
        self.ALT = [alt]
        self.end = end or pos
        self.gt_quals = gt_quals
        self.gt_types = gt_types
        self.var_type = var_type
        self.INFO = info_dict
        if var_id is None:
            var_id = "."
        self.ID = var_id


@pytest.fixture(scope="function")
def real_db_name(request):
    """Return the name of the real mongo test db"""
    return REAL_DATABASE


@pytest.fixture(scope="function")
def mongo_client(request):
    """Return a mongomock client"""
    client = MockClient(directConnection=True)
    return client


@pytest.fixture(scope="function")
def real_mongo_client(request):
    """Return a mongomock client"""
    client = MongoClient(directConnection=True)

    def teardown():
        print("\n")
        logger.info("Deleting database")
        client.drop_database(REAL_DATABASE)
        logger.info("Database deleted")

    request.addfinalizer(teardown)

    return client


@pytest.fixture(scope="function")
def mongo_adapter(request, mongo_client):
    """Return a mongo adapter"""
    db_name = TEST_DATABASE
    adapter = MongoAdapter(mongo_client, db_name)

    return adapter


@pytest.fixture(scope="function")
def real_mongo_adapter(request, real_mongo_client):
    """Return a mongo adapter"""
    db_name = REAL_DATABASE
    adapter = MongoAdapter(real_mongo_client, db_name)

    return adapter


@pytest.fixture(scope="function")
def simplest_variant(request):
    variant = {
        "_id": "test",
    }
    return variant


@pytest.fixture(scope="function")
def homozygous_variant(request):
    variant = {"_id": "test", "homozygote": 1, "case_id": "1"}
    return variant


@pytest.fixture(scope="function")
def simple_case(request):
    case = Case("test", "./test.vcf")

    return case


@pytest.fixture(scope="function")
def vcf_path(request):
    file_path = "tests/fixtures/test.vcf"
    return file_path


@pytest.fixture(scope="function")
def sv_vcf_path(request):
    file_path = "tests/fixtures/test.SV.vcf"
    return file_path


@pytest.fixture(scope="function")
def double_vcf_path(request):
    "This will return a vcf with the same variant two times"
    file_path = "tests/fixtures/double_variant.vcf"
    return file_path


@pytest.fixture(scope="function")
def unsorted_vcf_path(request):
    "This will return a vcf with unsorted variants"
    file_path = "tests/fixtures/unsorted.vcf"
    return file_path


@pytest.fixture(scope="function")
def zipped_vcf_path(request):
    "Returns a VCF that is gzipped and have a index companion"
    file_path = "tests/fixtures/test.vcf.gz"
    return file_path


@pytest.fixture(scope="function")
def profile_vcf_path(request):
    "Returns a VCF containing profile variants"
    file_path = "tests/fixtures/profile_snv.vcf"
    return file_path


@pytest.fixture(scope="function")
def ped_path(request):
    file_path = "tests/fixtures/recessive_trio.ped"
    return file_path


@pytest.fixture(scope="function")
def funny_ped_path(request):
    "Returns a ped file with incorrect family relations"
    return "tests/fixtures/funny_trio.ped"


@pytest.fixture(scope="function")
def ind_positions(request):
    return {"proband": 0, "mother": 1, "father": 2}


@pytest.fixture(scope="function")
def case_lines(request, ped_path):
    """Return ped formated case lines"""
    case = []
    with open(ped_path, "r") as f:
        for line in f:
            case.append(line)

    return case


@pytest.fixture(scope="function")
def case_id(request, ped_path):
    """Return ped formated case lines"""
    case_id = None
    with open(ped_path, "r") as f:
        for line in f:
            case_id = line.split("\t")[0]

    return case_id


@pytest.fixture(scope="function")
def vcf_obj(request, vcf_path):
    """return a cyvcf2.VCF obj"""
    return cyvcf2.VCF(vcf_path)


@pytest.fixture(scope="function")
def sv_vcf_obj(request, sv_vcf_path):
    """return a cyvcf2.VCF obj"""
    return cyvcf2.VCF(sv_vcf_path)


@pytest.fixture(scope="function")
def profile_list(request):
    return ["AA", "CC", "GG", "TT"]


@pytest.fixture(scope="function")
def case_obj(request, case_lines, vcf_obj, vcf_path, profile_list):
    """Return a case obj"""
    family_parser = FamilyParser(case_lines, family_type="ped")
    families = list(family_parser.families.keys())
    family = family_parser.families[families[0]]
    vcf_individuals = vcf_obj.samples
    nr_variants = 0
    for nr_variants, variant in enumerate(vcf_obj, 1):
        continue
    return build_case(
        case=family,
        vcf_individuals=vcf_individuals,
        vcf_path=vcf_path,
        nr_variants=nr_variants,
        profiles={individual: profile_list for individual in vcf_individuals},
    )


@pytest.fixture(scope="function")
def sv_case_obj(request, case_lines, sv_vcf_obj, sv_vcf_path):
    """Return a case obj"""
    family_parser = FamilyParser(case_lines, family_type="ped")
    families = list(family_parser.families.keys())
    family = family_parser.families[families[0]]
    vcf_individuals = sv_vcf_obj.samples
    nr_variants = 0
    for nr_variants, variant in enumerate(sv_vcf_obj, 1):
        continue
    return build_case(
        case=family,
        sv_individuals=vcf_individuals,
        vcf_sv_path=sv_vcf_path,
        nr_sv_variants=nr_variants,
    )


@pytest.fixture(scope="function")
def complete_case_obj(request, case_obj, sv_case_obj):
    """Return a case obj with both sv and snv information"""
    return update_case(case_obj, sv_case_obj)


@pytest.fixture(scope="function")
def family_obj(request, case_lines):
    """Return a case obj"""
    family_parser = FamilyParser(case_lines, family_type="ped")
    families = list(family_parser.families.keys())
    return family_parser.families[families[0]]


@pytest.fixture(scope="function")
def case_id(request, case_lines):
    """Return a case id"""
    family_parser = FamilyParser(case_lines, family_type="ped")
    families = list(family_parser.families.keys())
    family = family_parser.families[families[0]]
    return family.family_id


@pytest.fixture(scope="function")
def individuals(request, case_obj):
    """Return a case obj"""

    return case_obj.individuals


@pytest.fixture(scope="function")
def two_cases(request):
    """Return ped formated case lines"""
    return [
        "#FamilyID\tSampleID\tFather\tMother\tSex\tPhenotype\n",
        "1\tproband\t0\t0\t1\t2\n",
        "2\tproband\t0\t0\t2\t1\n",
    ]


## Variant fixtures:

# chrom='1',
# pos=80000,
# ref='A',
# alt='C',
# end=None,
# gt_quals=[60, 60, 60],
# gt_types=[1, 1, 0],
# var_type='snv',
# info_dict={}

### Variant objects


@pytest.fixture(scope="function")
def variant_obj(request, het_variant, ind_positions, individuals):
    return build_variant(
        variant=het_variant,
        individuals=individuals,
        ind_positions=ind_positions,
        case_id="test",
        gq_threshold=None,
    )


### CYVCF2 variants
### SNVs:
@pytest.fixture(scope="function")
def hem_variant(request):
    return CyvcfVariant(chrom="X", pos=60000)


@pytest.fixture(scope="function")
def variant_chr(request):
    return CyvcfVariant(chrom="chrX", pos=9000)


@pytest.fixture(scope="function")
def par_variant(request):
    return CyvcfVariant(chrom="X", pos=60001)


@pytest.fixture(scope="function")
def het_variant(request):
    return CyvcfVariant()


@pytest.fixture(scope="function")
def bnd_variant(request):
    return CyvcfVariant(
        chrom="1",
        pos=80000,
        alt="N[1:80000[",
        ref="N",
        end=80000,
        info_dict={"SVTYPE": "BND"},
    )


@pytest.fixture(scope="function")
def variant_no_gq(request):
    return CyvcfVariant(gt_quals=[-1, -1, -1])


@pytest.fixture(scope="function")
def hom_variant(request):
    return CyvcfVariant(gt_types=[3, 1, 1])


@pytest.fixture(scope="function")
def variant_no_call(request):
    return CyvcfVariant(gt_types=[2, 2, 2])


### SVs:
@pytest.fixture(scope="function")
def del_variant(request):
    return CyvcfVariant(
        chrom="1",
        ref="G",
        alt="<DEL>",
        pos=1285001,
        end=1287000,
        var_type="sv",
        info_dict={"END": 1287000, "SVLEN": -20000, "SVTYPE": "DEL"},
    )


@pytest.fixture(scope="function")
def small_insert_variant(request):
    return CyvcfVariant(
        chrom="1",
        ref="G",
        alt="GGGACGGGGGTTCTGAGATAAGCAAGCCCCCACCAGGTGAGACCGGCGGAGCTGTGGCCACCGAGGTCCCGGGAGCTGGTGCT",
        pos=3021145,
        end=3021145,
        var_type="sv",
        info_dict={"END": 3021145, "SVLEN": 82, "SVTYPE": "INS"},
    )


@pytest.fixture(scope="function")
def insertion_variant(request):
    return CyvcfVariant(
        chrom="1",
        ref="A",
        alt="<INS>",
        pos=3177306,
        end=3177306,
        var_type="sv",
        info_dict={"END": 3177306, "SVLEN": None, "SVTYPE": "INS"},
    )


@pytest.fixture(scope="function")
def duptandem_variant(request):
    return CyvcfVariant(
        chrom="1",
        ref="A",
        alt="<DUP:TANDEM>",
        pos=3092626,
        end=3092849,
        var_type="sv",
        info_dict={"END": 3092849, "SVLEN": 223, "SVTYPE": "DUP"},
    )


@pytest.fixture(scope="function")
def translocation_variant(request):
    return CyvcfVariant(
        chrom="1",
        ref="N",
        alt="N[11:119123896[",
        pos=3754913,
        end=3754913,
        var_type="sv",
        info_dict={"END": None, "SVLEN": None, "SVTYPE": "BND"},
    )


##SVs with chr prefix
@pytest.fixture(scope="function")
def chr_del_variant(request):
    return CyvcfVariant(
        chrom="chr1",
        ref="G",
        alt="<DEL>",
        pos=1285001,
        end=1287000,
        var_type="sv",
        info_dict={"END": 1287000, "SVLEN": -20000, "SVTYPE": "DEL"},
    )


@pytest.fixture(scope="function")
def chr_small_insert_variant(request):
    return CyvcfVariant(
        chrom="chr1",
        ref="G",
        alt="GGGACGGGGGTTCTGAGATAAGCAAGCCCCCACCAGGTGAGACCGGCGGAGCTGTGGCCACCGAGGTCCCGGGAGCTGGTGCT",
        pos=3021145,
        end=3021145,
        var_type="sv",
        info_dict={"END": 3021145, "SVLEN": 82, "SVTYPE": "INS"},
    )


@pytest.fixture(scope="function")
def chr_insertion_variant(request):
    return CyvcfVariant(
        chrom="chr1",
        ref="A",
        alt="<INS>",
        pos=3177306,
        end=3177306,
        var_type="sv",
        info_dict={"END": 3177306, "SVLEN": None, "SVTYPE": "INS"},
    )


@pytest.fixture(scope="function")
def chr_duptandem_variant(request):
    return CyvcfVariant(
        chrom="chr1",
        ref="A",
        alt="<DUP:TANDEM>",
        pos=3092626,
        end=3092849,
        var_type="sv",
        info_dict={"END": 3092849, "SVLEN": 223, "SVTYPE": "DUP"},
    )


@pytest.fixture(scope="function")
def chr_translocation_variant(request):
    return CyvcfVariant(
        chrom="chr1",
        ref="N",
        alt="N[11:119123896[",
        pos=3754913,
        end=3754913,
        var_type="sv",
        info_dict={"END": None, "SVLEN": None, "SVTYPE": "BND"},
    )



import pytest

from loqusdb.build_models.case import build_case, get_individual_positions
from loqusdb.exceptions import CaseError


def test_get_individual_positions():
    ## GIVEN a list with ids
    inds = ["1", "2", "3"]
    ## WHEN getting the individual positions
    ind_pos = get_individual_positions(inds)
    ## THEN assert they where given the correct position
    assert ind_pos["1"] == 0
    assert ind_pos["2"] == 1
    assert ind_pos["3"] == 2


def test_get_individual_positions_no_inds():
    ## GIVEN a list with ids
    inds = None
    ## WHEN getting the individual positions
    ind_pos = get_individual_positions(inds)
    ## THEN assert an empty dict is returned
    assert ind_pos == {}


def test_build_case_no_ped():
    ## GIVEN some vcf individuals

    vcf_individuals = ["mother", "proband"]
    case_id = "test"

    ## WHEN building a case object
    case_obj = build_case(
        case=None,
        vcf_individuals=vcf_individuals,
        case_id=case_id,
    )

    ## THEN assert that the case got the right ID
    assert case_obj["case_id"] == case_id
    for ind_obj in case_obj["individuals"]:
        assert ind_obj["name"] in vcf_individuals
        assert ind_obj["ind_id"] in vcf_individuals


def test_build_case_no_ped_no_case_id():
    ## WHEN building a case object

    ## THEN assert a CaseError is raised
    with pytest.raises(CaseError):
        ## GIVEN some vcf individuals

        vcf_individuals = ["mother", "proband"]

        case_obj = build_case(
            case=None,
            vcf_individuals=vcf_individuals,
        )


def test_build_case_ped(family_obj, vcf_path):
    ## GIVEN a ped parser family_obj
    vcf_inds = [ind_id for ind_id in family_obj.individuals]
    nr_variants = 10

    ## WHEN building a case object
    case_obj = build_case(
        case=family_obj,
        vcf_individuals=vcf_inds,
        vcf_path=vcf_path,
        nr_variants=nr_variants,
    )

    ## THEN assert that the case has the correct id
    assert case_obj["case_id"] == family_obj.family_id

    for ind_obj in case_obj["individuals"]:
        assert ind_obj["ind_id"] in vcf_inds

    ## THEN assert that the vcf_path was added
    assert case_obj["vcf_path"] == vcf_path

    ## THEN assert that the nr variants is correct
    assert case_obj["nr_variants"] == nr_variants


def test_build_case_ped_sv(family_obj, sv_vcf_path):
    ## GIVEN a ped parser family_obj
    vcf_inds = [ind_id for ind_id in family_obj.individuals]
    nr_sv_variants = 10

    ## WHEN building a case object
    case_obj = build_case(
        case=family_obj,
        sv_individuals=vcf_inds,
        vcf_sv_path=sv_vcf_path,
        nr_sv_variants=nr_sv_variants,
    )

    ## THEN assert that the case has the correct id
    assert case_obj["case_id"] == family_obj.family_id

    case_obj["individuals"] == []
    for ind_obj in case_obj["sv_individuals"]:
        assert ind_obj["ind_id"] in vcf_inds

    ## THEN assert that the vcf_path was added
    assert case_obj["vcf_path"] is None
    assert case_obj["vcf_sv_path"] == sv_vcf_path

    ## THEN assert that the nr variants is correct
    assert case_obj["nr_variants"] is None
    assert case_obj["nr_sv_variants"] == nr_sv_variants


def test_build_case_ped_sv_and_snv(family_obj, sv_vcf_path, vcf_path):
    ## GIVEN a ped parser family_obj
    vcf_inds = [ind_id for ind_id in family_obj.individuals]
    sv_vcf_inds = [ind_id for ind_id in family_obj.individuals]
    nr_sv_variants = 10
    nr_variants = 20

    ## WHEN building a case object
    case_obj = build_case(
        case=family_obj,
        sv_individuals=vcf_inds,
        vcf_sv_path=sv_vcf_path,
        nr_sv_variants=nr_sv_variants,
        vcf_individuals=vcf_inds,
        vcf_path=vcf_path,
        nr_variants=nr_variants,
    )

    ## THEN assert that the case has the correct id
    assert case_obj["case_id"] == family_obj.family_id

    for ind_obj in case_obj["individuals"]:
        assert ind_obj["ind_id"] in vcf_inds

    for ind_obj in case_obj["sv_individuals"]:
        assert ind_obj["ind_id"] in sv_vcf_inds

    ## THEN assert that the vcf_path was added
    assert case_obj["vcf_path"] == vcf_path
    assert case_obj["vcf_sv_path"] == sv_vcf_path

    ## THEN assert that the nr variants is correct
    assert case_obj["nr_variants"] == nr_variants
    assert case_obj["nr_sv_variants"] == nr_sv_variants



import logging

from loqusdb.plugins.mongo.structural_variant import SVMixin

from pymongo import ASCENDING, DESCENDING, DeleteOne, UpdateOne


LOG = logging.getLogger(__name__)


class VariantMixin(SVMixin):
    def _get_update(self, variant):
        """Convert a variant to a proper update

        Args:
            variant(dict)

        Returns:
            update(dict)
        """
        update = {
            "$inc": {
                "homozygote": variant.get("homozygote", 0),
                "hemizygote": variant.get("hemizygote", 0),
                "observations": 1,
            },
            "$set": {
                "chrom": variant.get("chrom"),
                "start": variant.get("pos"),
                "end": variant.get("end"),
                "ref": variant.get("ref"),
                "alt": variant.get("alt"),
            },
        }
        if variant.get("case_id"):
            update["$push"] = {"families": {"$each": [variant.get("case_id")], "$slice": -50}}
        return update

    def _get_update_delete(self, variant):
        update = {
            "$inc": {
                "homozygote": -variant.get("homozygote", 0),
                "hemizygote": -variant.get("hemizygote", 0),
                "observations": -1,
            },
            "$set": {
                "chrom": variant.get("chrom"),
                "start": variant.get("pos"),
                "end": variant.get("end"),
                "ref": variant.get("ref"),
                "alt": variant.get("alt"),
            },
        }
        if variant.get("case_id"):
            update["$pull"] = {"families": variant.get("case_id")}

        return update

    def add_variant(self, variant):
        """Add a variant to the variant collection

        If the variant exists we update the count else we insert a new variant object.

        Args:
            variant (dict): A variant dictionary

        """
        LOG.debug("Upserting variant: {0}".format(variant.get("_id")))

        update = self._get_update(variant)

        message = self.db.variant.update_one({"_id": variant["_id"]}, update, upsert=True)
        if message.modified_count == 1:
            LOG.debug("Variant %s was updated", variant.get("_id"))
        else:
            LOG.debug("Variant was added to database for first time")

    def add_variants(self, variants):
        """Add a bulk of variants

        This could be used for faster inserts

        Args:
            variants(iterable(dict))

        """

        operations = []
        nr_inserted = 0
        for i, variant in enumerate(variants, 1):
            # We need to check if there was any information returned
            # The variant could be excluded based on low gq or if no individiual was called
            # in the particular case
            if not variant:
                continue
            nr_inserted += 1
            update = self._get_update(variant)
            operations.append(UpdateOne({"_id": variant["_id"]}, update, upsert=True))
            if i % 10000 == 0:
                self.db.variant.bulk_write(operations, ordered=False)
                operations = []

        if operations:
            self.db.variant.bulk_write(operations, ordered=False)

        return nr_inserted

    def get_variant(self, variant):
        """Check if a variant exists in the database and return it.
        Variants are searched with a variant id which is a string that consists of
        chrom_pos_ref_alt.
        There is no simple way to create a similar ID for structural variants so this function
        will not work for SVs

        Search the variants with the variant id

        Args:
            variant(dict): A variant dictionary

        Returns:
            variant(dict): A variant dictionary or None
        """
        return self.db.variant.find_one({"_id": variant.get("_id")})

    def search_variants(self, variant_ids):
        """Make a batch search for variants in the database

        Args:
            variant_ids(list(str)): List of variant ids

        Returns:
            res(pymngo.Cursor(variant_obj)): The result
        """
        query = {"_id": {"$in": variant_ids}}

        return self.db.variant.find(query)

    def get_variants(self, chromosome=None, start=None, end=None):
        """Return all variants in the database
        If no region is specified all variants will be returned.

        Args:
            chromosome(str)
            start(int)
            end(int)


        Returns:
            variants(Iterable(Variant))
        """
        query = {}
        if chromosome:
            query["chrom"] = chromosome
        if start:
            query["start"] = {"$lte": end}
            query["end"] = {"$gte": start}
        LOG.info("Find all variants {}".format(query))
        return self.db.variant.find(query).sort([("start", ASCENDING)])

    def nr_variants(self, chromosome=None, start=None, end=None):
        """Return nr of variants"""

        query = {}
        if chromosome:
            query["chrom"] = chromosome
        if start:
            query["start"] = {"$lte": end}
            query["end"] = {"$gte": start}
        LOG.info("Find all variants {}".format(query))
        return self.db.variant.count_documents(query)

    def delete_variant(self, variant):
        """Delete observation in database

        This means that we take down the observations variable with one.
        If 'observations' == 1 we remove the variant. If variant was homozygote
        we decrease 'homozygote' with one.
        Also remove the family from array 'families'.

        Args:
            variant (dict): A variant dictionary

        """
        mongo_variant = self.get_variant(variant)

        if mongo_variant:

            if mongo_variant["observations"] == 1:
                LOG.debug("Removing variant {0}".format(mongo_variant.get("_id")))
                message = self.db.variant.delete_one({"_id": variant["_id"]})
            else:
                LOG.debug("Decreasing observations for {0}".format(mongo_variant.get("_id")))
                message = self.db.variant.update_one(
                    {"_id": mongo_variant["_id"]},
                    {
                        "$inc": {
                            "observations": -1,
                            "homozygote": -(variant.get("homozygote", 0)),
                            "hemizygote": -(variant.get("hemizygote", 0)),
                        },
                        "$pull": {"families": variant.get("case_id")},
                    },
                    upsert=False,
                )

    def delete_variants(self, variants):
        """Delete observations in database

        Given a list of variants, the write operation for each of the variant
        is given as a bulk to mongodb.

        Args:
            variants (list(Variant)): a list of variants
        """

        variant_id_dict = {variant["_id"]: variant for variant in variants}
        # Look up all variants at the same time to reduce number of operations
        # done on the database
        query = self.db.variant.find({"_id": {"$in": list(variant_id_dict.keys())}})
        operations = []
        for mongo_variant in query:
            variant = variant_id_dict.get(mongo_variant["_id"])
            if variant is None:
                continue
            if mongo_variant["observations"] == 1:
                operations.append(DeleteOne({"_id": variant["_id"]}))
                continue
            update = self._get_update_delete(variant)
            operations.append(UpdateOne({"_id": variant["_id"]}, update, upsert=False))
        # Make the accumulated write operations
        if operations:
            self.db.variant.bulk_write(operations, ordered=False)

    def get_chromosomes(self, sv=False):
        """Return a list of all chromosomes found in database

        Args:
            sv(bool): if sv variants should be choosen

        Returns:
            res(iterable(str)): An iterable with all chromosomes in the database
        """
        if sv:
            return self.db.structural_variant.distinct("chrom")
        else:
            return self.db.variant.distinct("chrom")

    def get_max_position(self, chrom):
        """Get the last position observed on a chromosome in the database

        Args:
            chrom(str)

        Returns:
            end(int): The largest end position found

        """
        res = (
            self.db.variant.find({"chrom": chrom}, {"_id": 0, "end": 1})
            .sort([("end", DESCENDING)])
            .limit(1)
        )
        end = 0
        for variant in res:
            end = variant["end"]
        return end



"""
loqusdb/utils/annotate.py

Functions to annotate variants in a VCF

The methods for annotating SNVs and SVs differs a bit so they will be handeld in separate.

"""

from loqusdb.build_models.variant import get_coords, get_variant_id


def annotate_variant(variant, var_obj=None):
    """Annotate a cyvcf variant with observations

    Args:
        variant(cyvcf2.variant)
        var_obj(dict)

    Returns:
        variant(cyvcf2.variant): Annotated variant
    """
    if var_obj:

        variant.INFO["Obs"] = var_obj["observations"]
        if var_obj.get("homozygote"):
            variant.INFO["Hom"] = var_obj["homozygote"]
        if var_obj.get("hemizygote"):
            variant.INFO["Hem"] = var_obj["hemizygote"]

    return variant


def annotate_svs(adapter, vcf_obj, keep_chr_prefix):
    """Annotate all SV variants in a VCF

    Args:
        adapter(loqusdb.plugin.adapter)
        vcf_obj(cyvcf2.VCF)

    Yields:
        variant(cyvcf2.Variant)
    """
    for nr_variants, variant in enumerate(vcf_obj, 1):
        variant_info = get_coords(variant, keep_chr_prefix)
        match = adapter.get_structural_variant(variant_info)
        if match:
            annotate_variant(variant, match)
        yield variant


def annotate_snvs(adapter, vcf_obj, keep_chr_prefix):
    """Annotate all variants in a VCF

    Args:
        adapter(loqusdb.plugin.adapter)
        vcf_obj(cyvcf2.VCF)

    Yields:
        variant(cyvcf2.Variant): Annotated variant
    """
    variants = {}

    for nr_variants, variant in enumerate(vcf_obj, 1):
        # Add the variant to current batch
        variants[get_variant_id(variant, keep_chr_prefix)] = variant
        # If batch len == 1000 we annotate the batch
        if (nr_variants % 1000) == 0:

            for var_obj in adapter.search_variants(list(variants.keys())):
                var_id = var_obj["_id"]
                if var_id in variants:
                    annotate_variant(variants[var_id], var_obj)

            for variant_id in variants:
                yield variants[variant_id]

            variants = {}

    for var_obj in adapter.search_variants(list(variants.keys())):
        var_id = var_obj["_id"]
        if var_id in variants:
            annotate_variant(variants[var_id], var_obj)

    for variant_id in variants:
        yield variants[variant_id]



from loqusdb.build_models.variant import build_variant
from loqusdb.constants import GRCH37, GRCH38


def test_get_insertion(small_insert_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a small insertion
    variant = small_insert_variant
    case_id = case_obj["case_id"]
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH37
    )

    adapter.add_case(case_obj)
    adapter.add_structural_variant(formated_variant)
    for variant_obj in adapter.db.structural_variant.find():
        assert variant_obj


def test_get_translocation(translocation_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a translocation
    variant = translocation_variant
    case_id = case_obj["case_id"]
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH37
    )

    adapter.add_case(case_obj)
    adapter.add_structural_variant(formated_variant)
    for variant_obj in adapter.db.structural_variant.find():
        assert variant_obj


def test_get_insertion_grch38(chr_small_insert_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a small insertion
    variant = chr_small_insert_variant
    case_id = case_obj["case_id"]
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH38
    )

    adapter.add_case(case_obj)
    adapter.add_structural_variant(formated_variant)
    for variant_obj in adapter.db.structural_variant.find():
        assert variant_obj


def test_get_translocation_grch38(chr_translocation_variant, mongo_adapter, case_obj):
    adapter = mongo_adapter
    ## GIVEN a mongo adapter with a translocation
    variant = chr_translocation_variant
    case_id = case_obj["case_id"]
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH38
    )

    adapter.add_case(case_obj)
    adapter.add_structural_variant(formated_variant)
    for variant_obj in adapter.db.structural_variant.find():
        assert variant_obj



import logging

import click

from loqusdb.commands.cli import cli as base_command

LOG = logging.getLogger(__name__)


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()


@base_command.command("wipe", short_help="Wipe a loqusdb instance")
@click.option(
    "--yes",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Are you sure you want to wipe the entire database?",
)
@click.pass_context
def wipe(ctx):
    """Wipe the entire db"""
    adapter = ctx.obj["adapter"]

    adapter.wipe_db()



import logging
import os
from datetime import datetime

import click
from loqusdb.utils.annotate import annotate_snvs, annotate_svs
from loqusdb.utils.vcf import add_headers, get_file_handle
from loqusdb.commands.cli import cli as base_command

LOG = logging.getLogger(__name__)


@base_command.command("annotate", short_help="Annotate a VCF with observations")
@click.argument(
    "variant-file",
    type=click.Path(exists=True),
    metavar="<vcf_file>",
)
@click.option("--sv", is_flag=True)
@click.pass_context
def annotate(ctx, variant_file, sv):
    """Annotate the variants in a VCF"""
    adapter = ctx.obj["adapter"]
    keep_chr_prefix = ctx.obj["keep_chr_prefix"]

    variant_path = os.path.abspath(variant_file)

    expected_type = "sv" if sv else "snv"

    nr_cases = adapter.nr_cases(sv_cases=True)
    LOG.info("Found {0} {1} cases in database".format(nr_cases, expected_type))

    vcf_obj = get_file_handle(variant_path)
    add_headers(vcf_obj, nr_cases=nr_cases, sv=sv)
    # Print the headers
    for header_line in vcf_obj.raw_header.split("\n"):
        if len(header_line) == 0:
            continue
        click.echo(header_line)

    start_inserting = datetime.now()

    if sv:
        annotated_variants = annotate_svs(adapter, vcf_obj, keep_chr_prefix)
    else:
        annotated_variants = annotate_snvs(adapter, vcf_obj, keep_chr_prefix)
    # try:
    for variant in annotated_variants:
        click.echo(str(variant).rstrip())



# -*- coding: utf-8 -*-
"""
loqusdb.utils.load.py

Functions to load data into the database.
This functions take an adapter which is the communication device for the database.

"""

import logging

import click

from loqusdb.build_models.case import build_case
from loqusdb.build_models.profile_variant import build_profile_variant
from loqusdb.build_models.variant import build_variant
from loqusdb.exceptions import CaseError, VcfError
from loqusdb.utils.case import get_case, update_case
from loqusdb.utils.delete import delete
from loqusdb.utils.profiling import get_profiles, profile_match
from loqusdb.utils.vcf import check_vcf, get_vcf

LOG = logging.getLogger(__name__)


def load_database(
    adapter,
    variant_file=None,
    sv_file=None,
    family_file=None,
    family_type="ped",
    skip_case_id=False,
    gq_threshold=None,
    snv_gq_only=False,
    keep_chr_prefix=False,
    qual_gq=False,
    case_id=None,
    max_window=3000,
    profile_file=None,
    hard_threshold=0.95,
    soft_threshold=0.9,
    genome_build=None,
    ignore_gq_if_unset=False,
):
    """Load the database with a case and its variants

    Args:
          adapter: Connection to database
          variant_file(str): Path to variant file
          sv_file(str): Path to sv variant file
          family_file(str): Path to family file
          family_type(str): Format of family file
          skip_case_id(bool): If no case information should be added to variants
          gq_threshold(int): If only quality variants should be considered
          keep_chr_prefix(bool): Retain chr/CHR/Chr prefix when present
          qual_gq(bool): Use QUAL field instead of GQ format tag to gate quality
          case_id(str): If different case id than the one in family file should be used
          max_window(int): Specify the max size for sv windows
          check_profile(bool): Does profile check if True
          hard_threshold(float): Rejects load if hamming distance above this is found
          soft_threshold(float): Stores similar samples if hamming distance above this is found
          genome_build(str): Store the genome version
          ignore_gq_if_unset(str): Ignore the gq threhsold check for variants that do not have a GQ or QUAL set

    Returns:
          nr_inserted(int)
    """
    vcf_files = []

    nr_variants = None
    vcf_individuals = None
    if variant_file:
        vcf_info = check_vcf(variant_file, keep_chr_prefix)
        nr_variants = vcf_info["nr_variants"]
        variant_type = vcf_info["variant_type"]
        vcf_files.append(variant_file)
        # Get the indivuduals that are present in vcf file
        vcf_individuals = vcf_info["individuals"]

    nr_sv_variants = None
    sv_individuals = None
    if sv_file:
        vcf_info = check_vcf(sv_file, keep_chr_prefix, "sv")
        nr_sv_variants = vcf_info["nr_variants"]
        vcf_files.append(sv_file)
        sv_individuals = vcf_info["individuals"]

    profiles = None
    matches = None
    if profile_file:
        profiles = get_profiles(adapter, profile_file, keep_chr_prefix)
        ###Check if any profile already exists
        matches = profile_match(
            adapter, profiles, hard_threshold=hard_threshold, soft_threshold=soft_threshold
        )

    # If a gq threshold is used the variants need to have GQ (only SNVs if snv_gq_only)
    for _vcf_file in vcf_files:
        is_sv = _vcf_file == sv_file
        if snv_gq_only and is_sv:
            continue  # skip GQ check for SV VCF

        vcf = get_vcf(_vcf_file)
        if gq_threshold and not vcf.contains("GQ") and not qual_gq:
            LOG.warning("Set gq-threshold to 0 or add info to vcf {0}".format(_vcf_file))
            raise SyntaxError("GQ is not defined in vcf header")

    # Get a ped_parser.Family object from family file
    family = None
    family_id = None
    if family_file:
        LOG.info("Loading family from %s", family_file)
        with open(family_file, "r") as family_lines:
            family = get_case(family_lines=family_lines, family_type=family_type)
            family_id = family.family_id

    # There has to be a case_id or a family at this stage.
    case_id = case_id or family_id
    # Convert infromation to a loqusdb Case object
    case_obj = build_case(
        case=family,
        case_id=case_id,
        vcf_path=variant_file,
        vcf_individuals=vcf_individuals,
        nr_variants=nr_variants,
        vcf_sv_path=sv_file,
        sv_individuals=sv_individuals,
        nr_sv_variants=nr_sv_variants,
        profiles=profiles,
        matches=matches,
        profile_path=profile_file,
    )
    # Build and load a new case, or update an existing one
    load_case(
        adapter=adapter,
        case_obj=case_obj,
    )

    nr_inserted = 0
    # If case was succesfully added we can store the variants
    for file_type in ["vcf_path", "vcf_sv_path"]:
        variant_type = "snv"
        if file_type == "vcf_sv_path":
            variant_type = "sv"
        if case_obj.get(file_type) is None:
            continue

        vcf_obj = get_vcf(case_obj[file_type])
        try:
            nr_inserted += load_variants(
                adapter=adapter,
                vcf_obj=vcf_obj,
                case_obj=case_obj,
                skip_case_id=skip_case_id,
                gq_threshold=gq_threshold if not snv_gq_only or variant_type == "snv" else None,
                qual_gq=qual_gq,
                keep_chr_prefix=keep_chr_prefix,
                max_window=max_window,
                variant_type=variant_type,
                genome_build=genome_build,
                ignore_gq_if_unset=ignore_gq_if_unset,
            )
        except Exception as err:
            # If something went wrong do a rollback
            LOG.warning(err)
            delete(
                adapter=adapter,
                case_obj=case_obj,
            )
            raise err
    return nr_inserted


def load_case(adapter, case_obj, update=False):
    """Load a case to the database

    Args:
        adapter: Connection to database
        case_obj: dict
        update(bool): If existing case should be updated

    Returns:
        case_obj(models.Case)
    """
    # Check if the case already exists in database.
    existing_case = adapter.case(case_obj)
    if existing_case:
        if not update:
            raise CaseError("Case {0} already exists in database".format(case_obj["case_id"]))
        case_obj = update_case(case_obj, existing_case)

    # Add the case to database

    adapter.add_case(case_obj, update=update)

    return case_obj


def load_variants(
    adapter,
    vcf_obj,
    case_obj,
    skip_case_id=False,
    gq_threshold=None,
    qual_gq=False,
    keep_chr_prefix=False,
    max_window=3000,
    variant_type="snv",
    genome_build=None,
    ignore_gq_if_unset=False,
):
    """Load variants for a family into the database.

    Args:
        adapter (loqusdb.plugins.Adapter): initialized plugin
        vcf_obj(cyvcf2.VCF): Iterable with cyvcf2.Variant
        case_obj(Case): dict with case information
        skip_case_id (bool): whether to include the case id on variant level
                             or not
        keep_chr_prefix(bool): Retain chr/CHR/Chr prefix when present
        gq_threshold(int)
        qual_gq(bool): whether to use QUAL instead of GQ
        max_window(int): Specify the max size for sv windows
        variant_type(str): 'sv' or 'snv'
        genome_build(str): Genome version. Ex. GRCH37
        ignore_gq_if_unset (bool): whether to add entries that have missing GQ or QUAL field

    Returns:
        nr_inserted(int)
    """
    if variant_type == "snv":
        nr_variants = case_obj["nr_variants"]
    else:
        nr_variants = case_obj["nr_sv_variants"]

    nr_inserted = 0
    case_id = case_obj["case_id"]
    if skip_case_id:
        case_id = None
    # Loop over the variants in the vcf
    with click.progressbar(vcf_obj, label="Inserting variants", length=nr_variants) as bar:

        variants = (
            build_variant(
                variant,
                case_obj,
                case_id,
                gq_threshold,
                qual_gq,
                keep_chr_prefix,
                ignore_gq_if_unset,
                genome_build=genome_build,
            )
            for variant in bar
        )

    if variant_type == "sv":
        for sv_variant in variants:
            if not sv_variant:
                continue
            adapter.add_structural_variant(variant=sv_variant, max_window=max_window)
            nr_inserted += 1

    if variant_type == "snv":
        nr_inserted = adapter.add_variants(variants)

    LOG.info("Inserted %s variants of type %s", nr_inserted, variant_type)

    return nr_inserted


def load_profile_variants(adapter, variant_file, keep_chr_prefix=None):
    """

    Loads variants used for profiling

    Args:
        adapter (loqusdb.plugins.Adapter): initialized plugin
        variant_file(str): Path to variant file


    """

    vcf_info = check_vcf(variant_file, keep_chr_prefix)
    variant_type = vcf_info["variant_type"]

    if variant_type != "snv":
        LOG.critical("Variants used for profiling must be SNVs only")
        raise VcfError

    vcf = get_vcf(variant_file)

    profile_variants = [build_profile_variant(variant, keep_chr_prefix) for variant in vcf]
    adapter.add_profile_variants(profile_variants)



import pytest
from loqusdb.exceptions import CaseError


def test_delete_family(mongo_adapter):
    ## GIVEN a mongo adapter with a case
    db = mongo_adapter.db
    case = {"case_id": "1", "vcf_path": "path_to_vcf"}

    db.case.insert_one(case)
    mongo_case = db.case.find_one()
    assert mongo_case["case_id"] == case["case_id"]

    ## WHEN deleting the case
    mongo_adapter.delete_case(case)

    ## THEN assert that the case was deleted
    mongo_case = db.case.find_one()

    assert mongo_case is None


def test_delete_non_existing_family(mongo_adapter):
    ## WHEN deleting a non existing case
    with pytest.raises(CaseError):
        ## GIVEN a mongo adapter and a empty database
        case = {"case_id": "1", "vcf_path": "path_to_vcf"}

        ## THEN assert a CaseError is raised
        mongo_adapter.delete_case(case)



import logging
import os
from datetime import datetime

import click
from loqusdb.exceptions import CaseError, VcfError
from loqusdb.utils.update import update_database

from loqusdb.commands.cli import cli as base_command

LOG = logging.getLogger(__name__)


@base_command.command("update", short_help="Update an existing case with a new type of variants")
@click.option(
    "--variant-file",
    type=click.Path(exists=True),
    metavar="<vcf_file>",
    help="Load a VCF with SNV/INDEL Variants",
)
@click.option(
    "--sv-variants",
    type=click.Path(exists=True),
    metavar="<sv_vcf_file>",
    help="Load a VCF with Structural Variants",
)
@click.option("-f", "--family-file", type=click.Path(exists=True), metavar="<ped_file>")
@click.option(
    "-t",
    "--family-type",
    type=click.Choice(["ped", "alt", "cmms", "mip"]),
    default="ped",
    show_default=True,
    help="If the analysis use one of the known setups, please specify which one.",
)
@click.option(
    "-c",
    "--case-id",
    type=str,
    help="If a different case id than the one in ped file should be used",
)
@click.option(
    "-s",
    "--skip-case-id",
    is_flag=True,
    show_default=True,
    help="Do not store case information on variants",
)
@click.option("--ensure-index", is_flag=True, help="Make sure that the indexes are in place")
@click.option("--gq-threshold", default=20, show_default=True, help="Threshold to consider variant")
@click.option(
    "--max-window",
    "-m",
    default=2000,
    show_default=True,
    help="Specify the maximum window size for svs",
)
@click.pass_context
def update(
    ctx,
    variant_file,
    sv_variants,
    family_file,
    family_type,
    skip_case_id,
    gq_threshold,
    case_id,
    ensure_index,
    max_window,
):
    """Load the variants of a case

    A variant is loaded if it is observed in any individual of a case
    If no family file is provided all individuals in vcf file will be considered.
    """
    if not (family_file or case_id):
        LOG.warning("Please provide a family file or a case id")
        ctx.abort()

    if not (variant_file or sv_variants):
        LOG.warning("Please provide a VCF file")
        ctx.abort()

    variant_path = None
    if variant_file:
        variant_path = os.path.abspath(variant_file)

    variant_sv_path = None
    if sv_variants:
        variant_sv_path = os.path.abspath(sv_variants)

    adapter = ctx.obj["adapter"]

    start_inserting = datetime.now()

    try:
        nr_inserted = update_database(
            adapter=adapter,
            variant_file=variant_path,
            sv_file=variant_sv_path,
            family_file=family_file,
            family_type=family_type,
            skip_case_id=skip_case_id,
            case_id=case_id,
            gq_threshold=gq_threshold,
            max_window=max_window,
        )
    except (SyntaxError, CaseError, IOError, VcfError) as error:
        LOG.warning(error)
        ctx.abort()

    LOG.info("Nr variants inserted: %s", nr_inserted)
    LOG.info("Time to insert variants: {0}".format(datetime.now() - start_inserting))

    if ensure_index:
        adapter.ensure_indexes()
    else:
        adapter.check_indexes()



from cyvcf2 import VCF

from loqusdb.utils.load import load_variants
from loqusdb.constants import GRCH37, GRCH38


def test_load_variants(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one heterozygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)
    mongo_variant = db.variant.find_one()

    assert mongo_variant == None

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH37)

    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 0
    assert mongo_variant["hemizygote"] == 0


def test_load_homozygote(real_mongo_adapter, hom_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one homozygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(hom_variant)
    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH37)
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 1
    assert mongo_variant["hemizygote"] == 0


def test_load_hemizygote(real_mongo_adapter, hem_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one hemizygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(hem_variant)

    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH37)
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 0
    assert mongo_variant["hemizygote"] == 1


def test_load_par_variant(real_mongo_adapter, par_variant, case_obj):
    mongo_adapter = real_mongo_adapter

    ## GIVEN an adapter and a vcf with one PAR variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(par_variant)

    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH37)
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 0
    assert mongo_variant["hemizygote"] == 0


def test_load_two_variants(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with tho heterygote variants
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)
    vcf.append(het_variant)

    ## WHEN loading the variants into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH37)

    ## THEN assert that the variant is loaded correct
    mongo_variant = db.variant.find_one()

    assert mongo_variant["observations"] == 2


def test_load_variants_skip_case_id(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with tho heterygote variants
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)

    ## WHEN loading the variants into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
        skip_case_id=True,
        genome_build=GRCH37,
    )

    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant.get("families") == None


def test_load_same_variant_different_case(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH37)

    case_id = case_obj["case_id"]
    case_id2 = "2"
    case_obj["case_id"] = case_id2

    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH37)

    mongo_variant = db.variant.find_one()

    assert mongo_variant["observations"] == 2
    assert mongo_variant["families"] == [case_id, case_id2]


def test_load_case_variants(real_mongo_adapter, case_obj):
    mongo_adapter = real_mongo_adapter

    db = mongo_adapter.db
    ## GIVEN a mongo adatper with snv variant file
    vcf_obj = VCF(case_obj["vcf_path"])
    ## WHEN loading the variants
    nr_variants = load_variants(
        adapter=mongo_adapter, vcf_obj=vcf_obj, case_obj=case_obj, genome_build=GRCH37
    )

    nr_loaded = 0
    for nr_loaded, variant in enumerate(db.variant.find(), 1):
        pass
    ## THEN assert that the correct number of variants was loaded
    assert nr_loaded > 0
    assert nr_loaded == case_obj["nr_variants"]


def test_load_sv_case_variants(mongo_adapter, sv_case_obj):
    db = mongo_adapter.db
    ## GIVEN a mongo adatper with snv variant file
    vcf_obj = VCF(sv_case_obj["vcf_sv_path"])
    ## WHEN loading the variants
    nr_variants = load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf_obj,
        case_obj=sv_case_obj,
        variant_type="sv",
        genome_build=GRCH37,
    )

    nr_loaded_svs = 0
    for nr_loaded_svs, variant in enumerate(db.structural_variant.find(), 1):
        pass
    nr_loaded_snvs = 0
    for nr_loaded_snvs, variant in enumerate(db.variant.find(), 1):
        pass
    ## THEN assert that the correct number of variants was loaded
    assert nr_loaded_svs > 0
    assert nr_loaded_snvs == 0
    assert nr_loaded_svs == sv_case_obj["nr_sv_variants"]


def test_load_variants_grch38(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one heterozygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)
    mongo_variant = db.variant.find_one()

    assert mongo_variant == None

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH38)

    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 0
    assert mongo_variant["hemizygote"] == 0


def test_load_homozygote_grch38(real_mongo_adapter, hom_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one homozygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(hom_variant)
    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH38)
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 1
    assert mongo_variant["hemizygote"] == 0


def test_load_hemizygote_grch38(real_mongo_adapter, variant_chr, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with one hemizygote variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(variant_chr)

    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 0
    assert mongo_variant["hemizygote"] == 1


def test_load_par_variant_grch38(real_mongo_adapter, par_variant, case_obj):
    mongo_adapter = real_mongo_adapter

    ## GIVEN an adapter and a vcf with one PAR variant
    db = mongo_adapter.db

    vcf = []
    vcf.append(par_variant)

    assert db.variant.find_one() == None

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH38)
    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant["families"] == [case_obj["case_id"]]
    assert mongo_variant["observations"] == 1
    assert mongo_variant["homozygote"] == 0
    assert mongo_variant["hemizygote"] == 0


def test_load_two_variants_grch38(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with tho heterygote variants
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)
    vcf.append(het_variant)

    ## WHEN loading the variants into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH38)

    ## THEN assert that the variant is loaded correct
    mongo_variant = db.variant.find_one()

    assert mongo_variant["observations"] == 2


def test_load_variants_skip_case_id_grch38(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf with tho heterygote variants
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)

    ## WHEN loading the variants into the database
    load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf,
        case_obj=case_obj,
        skip_case_id=True,
        genome_build=GRCH37,
    )

    mongo_variant = db.variant.find_one()

    ## THEN assert that the variant is loaded correct
    assert mongo_variant.get("families") == None


def test_load_same_variant_different_case_grch38(real_mongo_adapter, het_variant, case_obj):
    mongo_adapter = real_mongo_adapter
    ## GIVEN an adapter and a vcf
    db = mongo_adapter.db

    vcf = []
    vcf.append(het_variant)

    ## WHEN loading the variant into the database
    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH38)

    case_id = case_obj["case_id"]
    case_id2 = "2"
    case_obj["case_id"] = case_id2

    load_variants(adapter=mongo_adapter, vcf_obj=vcf, case_obj=case_obj, genome_build=GRCH38)

    mongo_variant = db.variant.find_one()

    assert mongo_variant["observations"] == 2
    assert mongo_variant["families"] == [case_id, case_id2]


def test_load_case_variants_grch38(real_mongo_adapter, case_obj):
    mongo_adapter = real_mongo_adapter

    db = mongo_adapter.db
    ## GIVEN a mongo adatper with snv variant file
    vcf_obj = VCF(case_obj["vcf_path"])
    ## WHEN loading the variants
    nr_variants = load_variants(
        adapter=mongo_adapter, vcf_obj=vcf_obj, case_obj=case_obj, genome_build=GRCH38
    )

    nr_loaded = 0
    for nr_loaded, variant in enumerate(db.variant.find(), 1):
        pass
    ## THEN assert that the correct number of variants was loaded
    assert nr_loaded > 0
    assert nr_loaded == case_obj["nr_variants"]


def test_load_sv_case_variants_grch38(mongo_adapter, sv_case_obj):
    db = mongo_adapter.db
    ## GIVEN a mongo adatper with snv variant file
    vcf_obj = VCF(sv_case_obj["vcf_sv_path"])
    ## WHEN loading the variants
    nr_variants = load_variants(
        adapter=mongo_adapter,
        vcf_obj=vcf_obj,
        case_obj=sv_case_obj,
        variant_type="sv",
        genome_build=GRCH38,
    )

    nr_loaded_svs = 0
    for nr_loaded_svs, variant in enumerate(db.structural_variant.find(), 1):
        pass
    nr_loaded_snvs = 0
    for nr_loaded_snvs, variant in enumerate(db.variant.find(), 1):
        pass
    ## THEN assert that the correct number of variants was loaded
    assert nr_loaded_svs > 0
    assert nr_loaded_snvs == 0
    assert nr_loaded_svs == sv_case_obj["nr_sv_variants"]



# -*- coding: utf-8 -*-
import logging
from datetime import datetime


from loqusdb.utils.vcf import get_vcf
from loqusdb.build_models.variant import build_variant

LOG = logging.getLogger(__name__)


def delete(
    adapter, case_obj, keep_chr_prefix=None, update=False, existing_case=False, genome_build=None
):
    """Delete a case and all of it's variants from the database.

    Args:
        adapter: Connection to database
        case_obj(models.Case)
        update(bool): If we are in the middle of an update
        existing_case(models.Case): If something failed during an update we need to revert
                                    to the original case
        keep_chr_prefix(bool): Retain chr/CHR/Chr prefixes in chromosome IDs when they are present
        genome_build(str): Genome build. Ex. GRCh37 or GRCh38

    """
    # This will overwrite the updated case with the previous one
    if update:
        adapter.add_case(existing_case)
    else:
        adapter.delete_case(case_obj)

    for file_type in ["vcf_path", "vcf_sv_path"]:
        if not case_obj.get(file_type):
            continue
        variant_file = case_obj[file_type]
        # Get a cyvcf2.VCF object
        vcf_obj = get_vcf(variant_file)

        if file_type == "vcf_path":
            LOG.info("deleting variants")
            delete_variants(
                adapter=adapter,
                vcf_obj=vcf_obj,
                keep_chr_prefix=keep_chr_prefix,
                case_obj=case_obj,
                genome_build=genome_build,
            )
        elif file_type == "vcf_sv_path":
            LOG.info("deleting structural variants")
            delete_structural_variants(
                adapter=adapter,
                vcf_obj=vcf_obj,
                case_obj=case_obj,
                keep_chr_prefix=keep_chr_prefix,
                genome_build=genome_build,
            )


def delete_variants(
    adapter, vcf_obj, case_obj, keep_chr_prefix=None, case_id=None, genome_build=None
):
    """Delete variants for a case in the database

    Args:
        adapter(loqusdb.plugins.Adapter)
        vcf_obj(iterable(dict))
        ind_positions(dict)
        case_id(str)
        genome_build(str): Genome build. Ex. GRCh37 or GRCh38

    Returns:
        nr_deleted (int): Number of deleted variants
    """
    case_id = case_id or case_obj["case_id"]
    nr_deleted = 0
    start_deleting = datetime.now()
    chrom_time = datetime.now()
    current_chrom = None
    new_chrom = None

    variant_list = []
    for variant in vcf_obj:
        formated_variant = build_variant(
            variant=variant,
            case_obj=case_obj,
            keep_chr_prefix=keep_chr_prefix,
            case_id=case_id,
            genome_build=genome_build,
        )

        if not formated_variant:
            continue

        variant_list.append(formated_variant)

        new_chrom = formated_variant.get("chrom")
        # When there are enough variants in the variant list
        # They are passed to delete_variants
        if len(variant_list) == 10000:
            adapter.delete_variants(variant_list)
            variant_list.clear()

        nr_deleted += 1

        if not current_chrom:
            LOG.info("Start deleting chromosome {}".format(new_chrom))
            current_chrom = new_chrom
            chrom_time = datetime.now()
            continue

        if new_chrom != current_chrom:
            LOG.info("Chromosome {0} done".format(current_chrom))
            LOG.info(
                "Time to delete chromosome {0}: {1}".format(
                    current_chrom, datetime.now() - chrom_time
                )
            )
            LOG.info("Start deleting chromosome {0}".format(new_chrom))
            current_chrom = new_chrom

    if variant_list:
        adapter.delete_variants(variant_list)
        variant_list.clear()

    return nr_deleted


def delete_structural_variants(
    adapter, vcf_obj, case_obj, genome_build, keep_chr_prefix=None, case_id=None
):
    """Delete structural variants for a case in the database

    Args:
        adapter(loqusdb.plugins.Adapter)
        vcf_obj(iterable(dict))
        ind_positions(dict)
        case_id(str)
        genome_build(str): Genome build. Ex. GRCh37 or GRCh38

    Returns:
        nr_deleted (int): Number of deleted variants"""

    case_id = case_id or case_obj["case_id"]
    nr_deleted = 0
    start_deleting = datetime.now()
    chrom_time = datetime.now()
    current_chrom = None
    new_chrom = None

    for variant in vcf_obj:
        formated_variant = build_variant(
            variant=variant,
            case_obj=case_obj,
            case_id=case_id,
            genome_build=genome_build,
            keep_chr_prefix=keep_chr_prefix,
        )

        if not formated_variant:
            continue

        new_chrom = formated_variant.get("chrom")
        adapter.delete_structural_variant(formated_variant)
        nr_deleted += 1

        if not current_chrom:
            LOG.info("Start deleting chromosome {}".format(new_chrom))
            current_chrom = new_chrom
            chrom_time = datetime.now()
            continue

        if new_chrom != current_chrom:
            LOG.info("Chromosome {0} done".format(current_chrom))
            LOG.info(
                "Time to delete chromosome {0}: {1}".format(
                    current_chrom, datetime.now() - chrom_time
                )
            )
            LOG.info("Start deleting chromosome {0}".format(new_chrom))
            current_chrom = new_chrom

    return nr_deleted



import pytest
from loqusdb.exceptions import ProfileError
from loqusdb.utils.load import load_profile_variants, load_case
from loqusdb.utils.profiling import get_profiles, profile_match, compare_profiles, check_duplicates
from loqusdb.utils.vcf import check_vcf


def test_get_profiles(real_mongo_adapter, profile_vcf_path, zipped_vcf_path):
    # Load profile variants
    load_profile_variants(real_mongo_adapter, profile_vcf_path)

    vcf_info = check_vcf(zipped_vcf_path)

    # Get profiles from vcf
    profiles = get_profiles(real_mongo_adapter, zipped_vcf_path, False)

    # Assert that all individuals are included
    assert list(profiles.keys()) == vcf_info["individuals"]

    # Assert that profile strings are of same lengths
    for i, individual in enumerate(profiles.keys()):
        if i == 0:
            length = len(profiles[individual])
        assert len(profiles[individual]) == length


def test_profile_match(real_mongo_adapter, profile_vcf_path, profile_list, case_obj):
    # Load profile variants
    load_profile_variants(real_mongo_adapter, profile_vcf_path)

    # Load case having profiles profile_list
    load_case(real_mongo_adapter, case_obj)

    # Get profiles from vcf
    profiles = {"test_individual": profile_list}

    # Assert that error is raised
    with pytest.raises(ProfileError) as error:
        profile_match(real_mongo_adapter, profiles)


def test_check_duplicates(real_mongo_adapter, profile_vcf_path, profile_list, case_obj):
    # Load profile variants
    load_profile_variants(real_mongo_adapter, profile_vcf_path)
    # Load case having profiles profile_list
    load_case(real_mongo_adapter, case_obj)
    # Create profiles dictionary
    profiles = {"test_individual": profile_list}
    # match profiles to the profiles in the database
    match = check_duplicates(real_mongo_adapter, profiles, hard_threshold=0.95)
    # This should match with the sample in the database
    assert match["profile"] == profile_list

    # Change last genotype, now no matches should be found
    profiles = {"test_individual": profile_list[:-1] + ["NN"]}
    match = check_duplicates(real_mongo_adapter, profiles, hard_threshold=0.80)
    assert match is None

    # Lower threshold. Now match should be found
    match = check_duplicates(real_mongo_adapter, profiles, hard_threshold=0.75)
    assert match["profile"] == profile_list


def test_compare_profiles():
    assert compare_profiles(["AA", "CC"], ["AA", "CC"]) == 1.0
    assert compare_profiles(["AA", "CC"], ["GG", "CC"]) == 0.5
    assert compare_profiles(["AC", "CG"], ["TC", "CG"]) == 0.5
    assert compare_profiles(["AC", "GT"], ["AA", "GG"]) == 0.0



# -*- coding: utf-8 -*-
import json
import logging
from pprint import pprint as pp

import click
from pymongo.cursor import Cursor

from loqusdb.commands.cli import cli as base_command

LOG = logging.getLogger(__name__)


@base_command.command("cases", short_help="Display cases in database")
@click.option("-c", "--case-id", help="Search for case")
@click.option("--to-json", is_flag=True)
@click.option("--count", is_flag=True, help="Show number of cases in database")
@click.option(
    "--case-type", "-t", help="Specify the type of cases", type=click.Choice(["snv", "sv"])
)
@click.pass_context
def cases(ctx, case_id, to_json, count, case_type):
    """Display cases in the database."""

    adapter = ctx.obj["adapter"]
    cases = []

    if count:
        snv_cases = None
        sv_cases = None
        if case_type == "snv":
            snv_cases = True
        elif case_type == "sv":
            sv_cases = True
        click.echo(adapter.nr_cases(snv_cases=snv_cases, sv_cases=sv_cases))
        return

    if case_id:
        case_obj = adapter.case({"case_id": case_id})
        if not case_obj:
            LOG.info("Case {0} does not exist in database".format(case_id))
            return
        case_obj["_id"] = str(case_obj["_id"])
        cases.append(case_obj)
    else:
        case_count: int = adapter.case_count()
        if case_count == 0:
            LOG.info("No cases found in database")
            ctx.abort()
        cases: Cursor = adapter.cases()

    if to_json:
        click.echo(json.dumps(cases))
        return

    click.echo("#case_id\tvcf_path")

    for case_obj in cases:
        case_id = case_obj.get("case_id")
        vcf_path = case_obj.get("vcf_path", case_obj.get("vcf_sv_path"))
        click.echo("{0}\t{1}".format(case_id, vcf_path))


@base_command.command("variants", short_help="Display variants in database")
@click.option("--variant-id", help="Search for a variant")
@click.option("-c", "--chromosome", help="Search for all variants in a chromosome")
@click.option("--end-chromosome", help="Search for all variants that ends on chromosome")
@click.option("-s", "--start", help="Start of region", type=int)
@click.option("-e", "--end", help="End of region", type=int)
@click.option(
    "-t",
    "--variant-type",
    help="Variant type to search for",
    type=click.Choice(["sv", "snv"]),
    default="snv",
)
@click.option(
    "--sv-type",
    help="Type of svs to search for",
)
@click.option("--to-json", is_flag=True)
@click.option("--case-count", is_flag=True, help="If number of cases should be included")
# @click.option('--sort-key',
#                 help='Specify what field to sort on',
# )
@click.pass_context
def variants(
    ctx,
    variant_id,
    chromosome,
    end_chromosome,
    start,
    end,
    variant_type,
    sv_type,
    to_json,
    case_count,
):
    """Display variants in the database."""
    if sv_type:
        variant_type = "sv"

    adapter = ctx.obj["adapter"]
    variant = {}
    if start or end:
        if not (chromosome and start and end):
            LOG.warning("Regions must be specified with chromosome, start and end")
            return

    nr_cases = None
    if case_count:
        snv_cases = None
        sv_cases = None
        if variant_type == "snv":
            snv_cases = True
        if variant_type == "sv":
            sv_cases = True
        nr_cases = adapter.nr_cases(snv_cases=snv_cases, sv_cases=sv_cases)

    if chromosome:
        if chromosome.startswith("chr"):
            chromosomes = [chromosome, chromosome[3:]]
        elif not chromosome.startswith("chr"):
            chromosomes = [chromosome, "chr" + chromosome]

    if end_chromosome:
        if end_chromosome.startswith("chr"):
            end_chromosomes = [end_chromosome, end_chromosome[3:]]
        elif not end_chromosome.startswith("chr"):
            end_chromosomes = [end_chromosome, "chr" + end_chromosome]
    else:
        end_chromosomes = [None, None]

    if variant_id:
        if variant_type == "sv":
            variant_query = {
                "chrom": chromosomes[0],
                "end_chrom": end_chromosomes[0] or chromosomes[0],
                "sv_type": sv_type,
                "pos": start,
                "end": end,
            }
            variant = list(adapter.get_structural_variant(variant_query))
            if len(variant) == 0:
                variant_query = {
                    "chrom": chromosomes[1],
                    "end_chrom": end_chromosomes[1] or chromosomes[1],
                    "sv_type": sv_type,
                    "pos": start,
                    "end": end,
                }
                variant = adapter.get_structural_variant(variant_query)

        else:
            variant = adapter.get_variant({"_id": variant_id})

        if not variant:
            LOG.info("Variant {0} does not exist in database".format(variant_id))

        if case_count:
            variant["total"] = nr_cases

        if to_json:
            LOG.info("Print in json format")
            if "_id" in variant:
                variant.pop("_id")
            click.echo(json.dumps(variant))
            return

        click.echo(variant)
        return
    if variant_type == "snv":
        result = list(adapter.get_variants(chromosome=chromosomes[0], start=start, end=end))
        if len(result) == 0:
            result = adapter.get_variants(chromosome=chromosomes[1], start=start, end=end)
    else:
        LOG.info("Search for svs")
        result = list(
            adapter.get_sv_variants(
                chromosome=chromosomes[0],
                end_chromosome=end_chromosomes[0],
                sv_type=sv_type,
                pos=start,
                end=end,
            )
        )
        if len(result) == 0:
            result = adapter.get_sv_variants(
                chromosome=chromosomes[1],
                end_chromosome=end_chromosomes[1],
                sv_type=sv_type,
                pos=start,
                end=end,
            )

    if to_json:
        json.dumps(variant)

    i = 0
    for i, variant in enumerate(result, 1):
        pp(variant)

    LOG.info("Number of variants found in database: %s", i)


@base_command.command("index", short_help="Add indexes to database")
@click.option(
    "--view",
    is_flag=True,
    help="Only display existing indexes",
)
@click.pass_context
def index(ctx, view):
    """Index the database."""
    adapter = ctx.obj["adapter"]
    if view:
        click.echo(adapter.indexes())
        return
    adapter.ensure_indexes()



import logging

from loqusdb.build_models.variant import get_variant_id

from loqusdb.models import ProfileVariant


LOG = logging.getLogger(__name__)


def get_maf(variant):
    """
    Gets the MAF (minor allele frequency) tag from the info field for the
    variant.

    Args:
        variant (cyvcf2.Variant)

    Returns:
        maf (float): Minor allele frequency

    """

    return variant.INFO.get("MAF")


def build_profile_variant(variant, keep_chr_prefix=None):
    """Returns a ProfileVariant object

    Args:
        variant (cyvcf2.Variant)
        keep_chr_prefix(bool): Retain chr/CHR/Chr prefix when present

    Returns:
        variant (models.ProfileVariant)
    """

    chrom = variant.CHROM
    if not keep_chr_prefix:
        if chrom.startswith(("chr", "CHR", "Chr")):
            chrom = chrom[3:]

    pos = int(variant.POS)

    variant_id = get_variant_id(variant, keep_chr_prefix)

    ref = variant.REF
    alt = variant.ALT[0]

    maf = get_maf(variant)

    profile_variant = ProfileVariant(
        variant_id=variant_id, chrom=chrom, pos=pos, ref=ref, alt=alt, maf=maf, id_column=variant.ID
    )

    return profile_variant



from loqusdb.build_models.variant import build_variant, GENOTYPE_MAP
from loqusdb.constants import GRCH37, GRCH38


def test_format_variant(het_variant, case_obj):
    ## GIVEN a parsed variant
    variant = het_variant
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH37
    )

    expected_id = "_".join([variant.CHROM, str(variant.POS), variant.REF, variant.ALT[0]])

    ## THEN assert it was built in a correct way
    assert formated_variant
    assert formated_variant["variant_id"] == expected_id
    assert formated_variant["chrom"] == variant.CHROM
    assert formated_variant["pos"] == variant.POS
    assert formated_variant["ref"] == variant.REF
    assert formated_variant["alt"] == variant.ALT[0]
    assert formated_variant["case_id"] == case_id
    assert formated_variant["homozygote"] == 0


def test_format_variant_chrprefix(het_variant, case_obj):
    ## GIVEN a parsed variant
    variant = het_variant
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )

    expected_id = "_".join([variant.CHROM, str(variant.POS), variant.REF, variant.ALT[0]])

    ## THEN assert it was built in a correct way
    assert formated_variant
    assert formated_variant["variant_id"] == expected_id
    assert formated_variant["chrom"] == variant.CHROM
    assert formated_variant["pos"] == variant.POS
    assert formated_variant["ref"] == variant.REF
    assert formated_variant["alt"] == variant.ALT[0]
    assert formated_variant["case_id"] == case_id
    assert formated_variant["homozygote"] == 0


def test_format_variant_no_qual(variant_no_gq, case_obj):
    ## GIVEN a variant without GQ
    variant = variant_no_gq
    ## And that has a missing QUAL value
    variant.QUAL = None
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant using a QUAL threshold
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        gq_qual=True,
        gq_threshold=20,
        genome_build=GRCH37,
    )
    ## THEN assert that None is returned since requirements are not fulfilled
    assert formated_variant is None


def test_format_variant_no_qual_chrprefix(variant_no_gq, case_obj):
    ## GIVEN a variant without GQ
    variant = variant_no_gq
    ## And that has a missing QUAL value
    variant.QUAL = None
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant using a QUAL threshold
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        gq_qual=True,
        gq_threshold=20,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )
    ## THEN assert that None is returned since requirements are not fulfilled
    assert formated_variant is None


def test_format_variant_no_gq(variant_no_gq, case_obj):
    ## GIVEN a variant without GQ
    variant = variant_no_gq
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant using a GQ threshold
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, gq_threshold=20, genome_build=GRCH37
    )
    ## THEN assert that None is returned since requirements are not fulfilled
    assert formated_variant is None


def test_format_variant_no_gq_chrprefix(variant_no_gq, case_obj):
    ## GIVEN a variant without GQ
    variant = variant_no_gq
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant using a GQ threshold
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        gq_threshold=20,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )
    ## THEN assert that None is returned since requirements are not fulfilled
    assert formated_variant is None


def test_format_variant_chr_prefix(variant_chr, case_obj):
    ## GIVEN a variant with 'chr' prefix in chromosome name
    variant = variant_chr
    assert variant.CHROM.startswith("chr")
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant using a GQ threshold
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, gq_threshold=20, genome_build=GRCH37
    )
    ## THEN assert that the 'chr' part has been stripped away
    assert formated_variant["chrom"] == variant.CHROM[3:]


def test_format_variant_chr_prefix_chrprefix(variant_chr, case_obj):
    ## GIVEN a variant with 'chr' prefix in chromosome name
    variant = variant_chr
    assert variant.CHROM.startswith("chr")
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant using a GQ threshold
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        gq_threshold=20,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )
    ## THEN assert that the 'chr' part has not been stripped away
    assert formated_variant["chrom"] == variant.CHROM


def test_format_variant_no_family_id(het_variant, case_obj):
    ## GIVEN a parsed variant
    variant = het_variant
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant telling that 'case_id' is None
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=None, genome_build=GRCH37
    )
    ## THEN assert that case_id was not added
    assert formated_variant.get("case_id") == None
    assert formated_variant["homozygote"] == 0
    assert formated_variant["hemizygote"] == 0


def test_format_variant_no_family_id_chrprefix(het_variant, case_obj):
    ## GIVEN a parsed variant
    variant = het_variant
    case_id = case_obj["case_id"]
    ## WHEN parsing the variant telling that 'case_id' is None
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=None, genome_build=GRCH38, keep_chr_prefix=True
    )
    ## THEN assert that case_id was not added
    assert formated_variant.get("case_id") == None
    assert formated_variant["homozygote"] == 0
    assert formated_variant["hemizygote"] == 0


def test_format_homozygote_variant(hom_variant, case_obj):
    ## GIVEN a parsed hom variant
    variant = hom_variant
    case_id = case_obj["case_id"]

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH37
    )

    ## THEN assert that the variant has hom count
    assert formated_variant["homozygote"] == 1
    assert formated_variant["hemizygote"] == 0


def test_format_homozygote_variant_chrprefix(hom_variant, case_obj):
    ## GIVEN a parsed hom variant
    variant = hom_variant
    case_id = case_obj["case_id"]

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )

    ## THEN assert that the variant has hom count
    assert formated_variant["homozygote"] == 1
    assert formated_variant["hemizygote"] == 0


def test_format_hemizygote_variant(hem_variant, case_obj):
    ## GIVEN a parsed hemizygous variant
    variant = hem_variant
    case_id = case_obj["case_id"]

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH37
    )

    ## THEN assert that hemizygote count is 1
    assert formated_variant["homozygote"] == 0
    assert formated_variant["hemizygote"] == 1


def test_format_hemizygote_variant_chrprefix(variant_chr, case_obj):
    ## GIVEN a parsed hemizygous variant
    variant = variant_chr
    case_id = case_obj["case_id"]

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )

    ## THEN assert that hemizygote count is 1
    assert formated_variant["homozygote"] == 0
    assert formated_variant["hemizygote"] == 1


def test_format_variant_no_call(variant_no_call, case_obj):
    ## GIVEN a parsed variant with no call in all individuals
    variant = variant_no_call
    case_id = case_obj["case_id"]

    for call in variant.gt_types:
        assert GENOTYPE_MAP[call] in ["no_call", "hom_ref"]

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant, case_obj=case_obj, case_id=case_id, genome_build=GRCH37
    )

    ## THEN assert that the result is None
    assert formated_variant is None


def test_format_variant_no_call_chrprefix(variant_no_call, case_obj):
    ## GIVEN a parsed variant with no call in all individuals
    variant = variant_no_call
    case_id = case_obj["case_id"]

    for call in variant.gt_types:
        assert GENOTYPE_MAP[call] in ["no_call", "hom_ref"]

    ## WHEN parsing the variant
    formated_variant = build_variant(
        variant=variant,
        case_obj=case_obj,
        case_id=case_id,
        genome_build=GRCH38,
        keep_chr_prefix=True,
    )

    ## THEN assert that the result is None
    assert formated_variant is None



import logging

from loqusdb.exceptions import CaseError

LOG = logging.getLogger(__name__)


class CaseMixin:
    def case(self, case):
        """Get a case from the database

        Search the cases with the case id

        Args:
            case (dict): A case dictionary

        Returns:
            mongo_case (dict): A mongo case dictionary
        """
        LOG.debug("Getting case {0} from database".format(case.get("case_id")))
        case_id = case["case_id"]
        return self.db.case.find_one({"case_id": case_id})

    def cases(self):
        """Get all cases from the database

        Returns:
            cases (Iterable(Case)): A iterable with mongo cases
        """
        LOG.debug("Collecting all cases from database")
        return self.db.case.find()

    def nr_cases(self, snv_cases=None, sv_cases=None):
        """Return the number of cases in the database

        Args:
            snv_cases(bool): If only snv cases should be searched
            sv_cases(bool): If only snv cases should be searched

        Returns:
            cases (Iterable(Case)): A iterable with mongo cases
        """
        LOG.info("Fetch number of cases")
        query = {}

        if snv_cases:
            LOG.info("Fetch all snv cases")
            query = {"vcf_path": {"$ne": None}}
        if sv_cases:
            LOG.info("Fetch all sv cases")
            query = {"vcf_sv_path": {"$ne": None}}
        if snv_cases and sv_cases:
            query = {}

        return self.db.case.count_documents(query)

    def add_case(self, case, update=False):
        """Add a case to the case collection

        If the case exists and update is False raise error.

        Args:
            db (MongoClient): A connection to the mongodb
            case (dict): A case dictionary
            update(bool): If existing case should be updated

        Returns:
            mongo_case_id(ObjectId)

        """
        existing_case = self.case(case)
        if existing_case and not update:
            raise CaseError("Case {} already exists".format(case["case_id"]))
        if existing_case:
            self.db.case.find_one_and_replace(
                {"case_id": case["case_id"]},
                case,
            )
        else:
            self.db.case.insert_one(case)

        return case

    def delete_case(self, case):
        """Delete case from the database

        Delete a case from the database

        Args:
            case (dict): A case dictionary

        """
        mongo_case = self.case(case)

        if not mongo_case:
            raise CaseError(
                "Tried to delete case {0} but could not find case".format(case.get("case_id"))
            )
        LOG.info("Removing case {0} from database".format(mongo_case.get("case_id")))
        self.db.case.delete_one({"_id": mongo_case["_id"]})

    def case_count(self) -> int:
        """Returns the total number of cases in the database."""
        return self.db.case.count_documents({})



from click.testing import CliRunner

from loqusdb.commands.cli import cli as base_command


def test_export_base(real_db_name: str):
    """Test the base command that exports variants."""

    runner = CliRunner()

    # WHEN the base command to export cases is run
    command = ["--database", real_db_name, "export"]

    ## THEN it should return success
    result = runner.invoke(base_command, command)
    assert result.exit_code == 0



import logging
from datetime import datetime

import click
from loqusdb.utils.migrate import migrate_database

from loqusdb.commands.cli import cli as base_command

LOG = logging.getLogger(__name__)


@base_command.command("migrate", short_help="Migrate an old loqusdb instance")
@click.pass_context
def migrate(
    ctx,
):
    """Migrate an old loqusdb instance to 1.0"""
    adapter = ctx.obj["adapter"]

    start_time = datetime.now()

    nr_updated = migrate_database(adapter)

    LOG.info(
        "All variants updated, time to complete migration: {}".format(datetime.now() - start_time)
    )
    LOG.info("Nr variants that where updated: %s", nr_updated)



from loqusdb.utils.delete import delete
from loqusdb.utils.load import load_database
from loqusdb.constants import GRCH37, GRCH38


def test_delete_case(mongo_adapter, simple_case):
    ## GIVEN a mongoadapter with a inserted case
    db = mongo_adapter.db
    db.case.insert_one(simple_case)
    mongo_case = db.case.find_one()

    assert mongo_case
    ## WHEN deleting the case

    mongo_adapter.delete_case(simple_case)
    ## THEN assert that the case was deleted
    mongo_case = db.case.find_one()

    mongo_case = db.case.find_one()

    assert mongo_case is None


def test_delete_case_and_variants(vcf_path, ped_path, real_mongo_adapter, case_id, case_obj):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type="ped",
        genome_build=GRCH37,
    )

    mongo_case = db.case.find_one()

    assert mongo_case["case_id"] == case_id

    delete(adapter=mongo_adapter, case_obj=case_obj, genome_build=GRCH37)

    mongo_case = db.case.find_one()

    assert mongo_case is None

    mongo_variant = db.variant.find_one()

    assert mongo_variant is None


def test_delete_structural_variants(vcf_path, ped_path, real_mongo_adapter, case_id, sv_case_obj):
    # GIVEN a mongo adapter with an inserted case with SVs
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        family_file=ped_path,
        family_type="ped",
        sv_file=sv_case_obj["vcf_sv_path"],
        genome_build=GRCH37,
    )

    mongo_svs = db.structural_variant.find()
    assert len(list(mongo_svs)) == 19

    # WHEN deleteing the case
    delete(adapter=mongo_adapter, case_obj=sv_case_obj, genome_build=GRCH37)

    # All structural variants should be deleted.
    mongo_svs = db.structural_variant.find()
    assert len(list(mongo_svs)) == 0


def test_delete_case_and_variants_grch38(vcf_path, ped_path, real_mongo_adapter, case_id, case_obj):
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type="ped",
        genome_build=GRCH37,
    )

    mongo_case = db.case.find_one()

    assert mongo_case["case_id"] == case_id

    delete(adapter=mongo_adapter, case_obj=case_obj, genome_build=GRCH38)

    mongo_case = db.case.find_one()

    assert mongo_case is None

    mongo_variant = db.variant.find_one()

    assert mongo_variant is None


def test_delete_structural_variants_grch38(
    vcf_path, ped_path, real_mongo_adapter, case_id, sv_case_obj
):
    # GIVEN a mongo adapter with an inserted case with SVs
    mongo_adapter = real_mongo_adapter
    db = mongo_adapter.db

    load_database(
        adapter=mongo_adapter,
        family_file=ped_path,
        family_type="ped",
        sv_file=sv_case_obj["vcf_sv_path"],
        genome_build=GRCH37,
    )

    mongo_svs = db.structural_variant.find()
    assert len(list(mongo_svs)) == 19

    # WHEN deleteing the case
    delete(adapter=mongo_adapter, case_obj=sv_case_obj, genome_build=GRCH38)

    # All structural variants should be deleted.
    mongo_svs = db.structural_variant.find()
    assert len(list(mongo_svs)) == 0



from click.testing import CliRunner

from loqusdb.commands.cli import cli as base_command


def test_identity(real_db_name: str):
    """Test the SV identity base command."""

    runner = CliRunner()

    # WHEN the base identity command is run on an empty database
    command = ["--database", real_db_name, "identity", "-v", "1_7890024_TGA_GGG"]

    # THEN the command should return success
    result = runner.invoke(base_command, command)
    assert result.exit_code == 0

    # AND no variant found message
    assert "No hits for variant" in result.output



import pkg_resources

from loqusdb.constants import GRCH37, GRCH38

###### Files ######

# Background data

backround_svs = "resources/loqusdb.20181005.gz"
maf_50_grch37 = "resources/maf_50_sites_GRCh37.vcf.gz"
maf_50_grch38 = "resources/maf_50_sites_GRCh38.vcf.gz"

### Profile SNPs for Grch37 and 38

###### Paths ######

# Backround data path

background_path = pkg_resources.resource_filename("loqusdb", backround_svs)

maf_grch37_path = pkg_resources.resource_filename("loqusdb", maf_50_grch37)
maf_grch38_path = pkg_resources.resource_filename("loqusdb", maf_50_grch38)

MAF_PATH = {GRCH37: maf_grch37_path, GRCH38: maf_grch38_path}



import logging
from copy import deepcopy

import numpy as np
from loqusdb.build_models.variant import get_variant_id
from loqusdb.constants import GENOTYPE_MAP, HAMMING_RANGES
from loqusdb.exceptions import ProfileError

from .vcf import get_file_handle

LOG = logging.getLogger(__name__)


def get_profiles(adapter, vcf_file, keep_chr_prefix):
    """Given a vcf, get a profile string for each sample in the vcf
    based on the profile variants in the database

    Args:
        adapter(MongoAdapter): Adapter to mongodb
        vcf_file(str): Path to vcf file

    Returns:
        profiles (dict(str)): The profiles (given as strings) for each sample
                              in vcf.

    """
    vcf = get_file_handle(vcf_file)
    individuals = vcf.samples
    profiles = {individual: [] for individual in individuals}

    for profile_variant in adapter.profile_variants():

        ref = profile_variant["ref"]
        alt = profile_variant["alt"]

        pos = profile_variant["pos"]
        end = pos + 1
        chrom = profile_variant["chrom"]

        region = f"{chrom}:{pos}-{end}"

        # Find variants in region

        found_variant = False
        for variant in vcf(region):

            variant_id = get_variant_id(variant, keep_chr_prefix)

            # If variant id i.e. chrom_pos_ref_alt matches
            if variant_id == profile_variant["_id"]:
                found_variant = True
                # find genotype for each individual in vcf
                for i, individual in enumerate(individuals):

                    genotype = GENOTYPE_MAP[variant.gt_types[i]]
                    if genotype == "hom_alt":
                        gt_str = f"{alt}{alt}"
                    elif genotype == "het":
                        gt_str = f"{ref}{alt}"
                    else:
                        gt_str = f"{ref}{ref}"

                    # Append genotype to profile string of individual
                    profiles[individual].append(gt_str)

                # Break loop if variant is found in region
                break

        # If no call was found for variant, give all samples a hom ref genotype
        if not found_variant:
            for individual in individuals:
                profiles[individual].append(f"{ref}{ref}")

    return profiles


def profile_match(adapter, profiles, hard_threshold=0.95, soft_threshold=0.9):
    """
    given a dict of profiles, searches through all the samples in the DB
    for a match. If a matching sample is found an exception is raised,
    and the variants will not be loaded into the database.

    Args:
        adapter (MongoAdapter): Adapter to mongodb
        profiles (dict(str)): The profiles (given as strings) for each sample in vcf.
        hard_threshold(float): Rejects load if hamming distance above this is found
        soft_threshold(float): Stores similar samples if hamming distance above this is found

    Returns:
        matches(dict(list)): list of similar samples for each sample in vcf.
    """
    matches = {sample: [] for sample in profiles.keys()}
    for case in adapter.cases():

        if case.get("individuals") is None:
            continue

        for individual in case["individuals"]:

            for sample in profiles.keys():

                if individual.get("profile"):

                    similarity = compare_profiles(profiles[sample], individual["profile"])

                    if similarity >= hard_threshold:
                        msg = (
                            f"individual {sample} has a {similarity} similarity "
                            f"with individual {individual['ind_id']} in case "
                            f"{case['case_id']}"
                        )
                        LOG.critical(msg)

                        # Raise some exception
                        raise ProfileError

                    if similarity >= soft_threshold:
                        match = f"{case['case_id']}.{individual['ind_id']}"
                        matches[sample].append(match)

    return matches


def check_duplicates(adapter, profiles, hard_threshold):
    """
    Searches database for duplicates. If duplicate is found, the individual
    is returned.

    Args:
        adapter (MongoAdapter): Adapter to mongodb
        profiles (dict(str)): The profiles (given as strings) for each sample in vcf.
        hard_threshold(float): Rejects load if hamming distance above this is found
    Returns:
        individual (dict): dictionary representation of duplicated individual

    """

    for case in adapter.cases():

        if case.get("individuals") is None:
            continue

        for individual in case["individuals"]:

            for sample in profiles.keys():

                if individual.get("profile"):

                    similarity = compare_profiles(profiles[sample], individual["profile"])

                    if similarity >= hard_threshold:
                        msg = (
                            f"individual {sample} has a {similarity} similarity "
                            f"with individual {individual['ind_id']} in case "
                            f"{case['case_id']}"
                        )
                        LOG.info(msg)
                        return individual


def compare_profiles(profile1, profile2):
    """
    Given two profiles, determine the ratio of similarity, i.e.
    the hamming distance between the strings.

    Args:
        profile1/2 (str): profile string
    Returns:
        similarity_ratio (float): the ratio of similiarity (0-1)
    """

    length = len(profile1)

    profile1 = np.array(list(profile1))
    profile2 = np.array(list(profile2))

    similarity_array = profile1 == profile2

    matches = np.sum(similarity_array)

    similarity_ratio = matches / length

    return similarity_ratio


def update_profiles(adapter, keep_chr_prefix):
    """
    For all cases having vcf_path, update the profile string for the samples

    Args:
        adapter (MongoAdapter): Adapter to mongodb

    """

    for case in adapter.cases():

        # If the case has a vcf_path, get the profiles and update the
        # case with new profiled individuals.
        if case.get("profile_path"):

            profiles = get_profiles(adapter, case["profile_path"], keep_chr_prefix)
            profiled_individuals = deepcopy(case["individuals"])

            for individual in profiled_individuals:
                ind_id = individual["ind_id"]
                try:
                    profile = profiles[ind_id]
                    individual["profile"] = profile

                except KeyError:
                    LOG.warning(f"sample IDs in vcf does not match for case {case['case_id']}")

            updated_case = deepcopy(case)

            updated_case["individuals"] = profiled_individuals

            adapter.add_case(updated_case, update=True)


def profile_stats(adapter, threshold=0.9):
    """
    Compares the pairwise hamming distances for all the sample profiles in
    the database. Returns a table of the number of distances within given
    ranges.

    Args:
        adapter (MongoAdapter): Adapter to mongodb
        threshold (float): If any distance is found above this threshold
            a warning will be given, stating the two matching samples.

    Returns:
        distance_dict (dict): dictionary with ranges as keys, and the number
            of distances that are within these ranges as values.

    """
    profiles = []
    samples = []

    # Instatiate the distance dictionary with a count 0 for all the ranges
    distance_dict = {key: 0 for key in HAMMING_RANGES.keys()}

    for case in adapter.cases():

        for individual in case["individuals"]:

            if individual.get("profile"):
                # Make sample name <case_id>.<sample_id>
                sample_id = f"{case['case_id']}.{individual['ind_id']}"
                ind_profile = individual["profile"]

                # Numpy array to hold all the distances for this samples profile
                distance_array = np.array([], dtype=np.float)

                for sample, profile in zip(samples, profiles):

                    # Get distance and append to distance array
                    distance = compare_profiles(ind_profile, profile)
                    distance_array = np.append(distance_array, distance)

                    # Issue warning if above threshold
                    if distance >= threshold:
                        LOG.warning(f"{sample_id} is {distance} similar to {sample}")

                # Check number of distances in each range and add to distance_dict
                for key, range in HAMMING_RANGES.items():
                    # Calculate the number of hamming distances found within the
                    # range for current individual
                    distance_dict[key] += np.sum(
                        (distance_array >= range[0]) & (distance_array < range[1])
                    )

                # Append profile and sample_id for this sample for the next
                # iteration
                profiles.append(ind_profile)
                samples.append(sample_id)

    return distance_dict



import pytest
from loqusdb.exceptions import CaseError

from loqusdb.utils.load import update_case, load_case


def test_load_family(mongo_adapter, simple_case):
    ## GIVEN a adapter to an empty database
    db = mongo_adapter.db

    ## WHEN loading a family to the database
    mongo_adapter.add_case(simple_case)

    ## THEN assert that the case was loaded
    mongo_case = db.case.find_one()
    assert mongo_case["case_id"] == simple_case["case_id"]


def test_load_same_family_twice(mongo_adapter, simple_case):
    ## GIVEN a adapter to an empty database
    db = mongo_adapter.db

    ## WHEN loading a family to the database twice
    mongo_adapter.add_case(simple_case)

    ## THEN assert that a CaseError is raised
    with pytest.raises(CaseError):
        mongo_adapter.add_case(simple_case)


def test_update_case(case_obj, sv_case_obj):
    ## GIVEN an existing case and a case with new info

    ## WHEN merging the two case objs
    updated_case = update_case(sv_case_obj, case_obj)

    ## THEN assert that the updated case includes information from both cases
    assert updated_case["sv_individuals"] == sv_case_obj["sv_individuals"]
    assert updated_case["sv_individuals"] != case_obj["sv_individuals"]

    assert updated_case["_sv_inds"] == sv_case_obj["_sv_inds"]
    assert updated_case["_sv_inds"] != case_obj["_sv_inds"]

    assert updated_case["nr_variants"] == case_obj["nr_variants"]
    assert updated_case["nr_sv_variants"] == sv_case_obj["nr_sv_variants"]


def test_update_case_same_vcf(case_obj, sv_case_obj):
    ## GIVEN an existing case and a case with new info
    sv_case_obj["vcf_path"] = case_obj["vcf_path"]
    ## WHEN merging the two case objs
    ## THEN assert that a CaseError is raised since we are trying to modify existing VCF
    with pytest.raises(CaseError):
        updated_case = update_case(sv_case_obj, case_obj)


def test_load_complete_case(mongo_adapter, complete_case_obj):
    ## GIVEN a case that includes both svs and snvs
    db = mongo_adapter.db

    ## WHEN loading the case
    case_obj = load_case(mongo_adapter, complete_case_obj)
    ## THEN assert that all info is added
    loaded_case = db.case.find_one()

    assert len(loaded_case["individuals"]) == 3
    assert len(loaded_case["sv_individuals"]) == 3
    assert loaded_case["nr_variants"] > 0
    assert loaded_case["nr_sv_variants"] > 0



import logging
from copy import deepcopy

from loqusdb.exceptions import CaseError
from ped_parser import FamilyParser

LOG = logging.getLogger(__name__)


def get_case(family_lines, family_type="ped", vcf_path=None):
    """Return ped_parser case from a family file

    Create a dictionary with case data. If no family file is given create from VCF

    Args:
        family_lines (iterator): The family lines
        family_type (str): The format of the family lines
        vcf_path(str): Path to VCF

    Returns:
        family (Family): A ped_parser family object
    """
    family = None
    LOG.info("Parsing family information")

    family_parser = FamilyParser(family_lines, family_type)

    families = list(family_parser.families.keys())

    LOG.info("Found families {0}".format(", ".join(families)))

    if len(families) > 1:
        raise CaseError("Only one family per load can be used")

    family = family_parser.families[families[0]]

    return family


def update_case(case_obj, existing_case):
    """Update an existing case

    This will add paths to VCF files, individuals etc

    Args:
        case_obj(models.Case)
        existing_case(models.Case)

    Returns:
        updated_case(models.Case): Updated existing case
    """
    variant_nrs = ["nr_variants", "nr_sv_variants"]
    individuals = [("individuals", "_inds"), ("sv_individuals", "_sv_inds")]

    updated_case = deepcopy(existing_case)

    for i, file_name in enumerate(["vcf_path", "vcf_sv_path"]):
        variant_type = "snv"
        if file_name == "vcf_sv_path":
            variant_type = "sv"
        if case_obj.get(file_name):
            if updated_case.get(file_name):
                LOG.warning("VCF of type %s already exists in case", variant_type)
                raise CaseError("Can not replace VCF in existing case")
            else:
                updated_case[file_name] = case_obj[file_name]
                updated_case[variant_nrs[i]] = case_obj[variant_nrs[i]]
                updated_case[individuals[i][0]] = case_obj[individuals[i][0]]
                updated_case[individuals[i][1]] = case_obj[individuals[i][1]]

    return updated_case



# -*- coding: utf-8 -*-


class Variant(dict):
    """Represent a Variant.

    snv/indels and svs differ quite a lot, we are here representing them with the same object.
    snv/indels will miss some of the information
    """

    def __init__(
        self,
        chrom,
        pos,
        end,
        ref,
        alt,
        variant_id=None,
        end_chrom=None,
        sv_type=None,
        sv_len=None,
        case_id=None,
        observations=0,
        homozygote=0,
        hemizygote=0,
        is_sv=False,
        id_column=None,
    ):
        super(Variant, self).__init__(
            _id=variant_id,
            variant_id=variant_id,
            chrom=chrom,
            pos=pos,
            end=end,
            ref=ref,
            alt=alt,
            end_chrom=end_chrom,
            sv_type=sv_type,
            sv_len=sv_len,
            case_id=case_id,
            observations=observations,
            homozygote=homozygote,
            hemizygote=hemizygote,
            is_sv=is_sv,
            id_column=id_column,
        )



from .annotate import annotate as annotate_command
from .cli import cli as base_command
from .delete import delete as delete_command
from .export import export as export_command
from .identity import identity as identity_command
from .load import load as load_command
from .load_profile import load_profile as profile_command
from .migrate import migrate as migration_command
from .restore import restore as restore_command
from .update import update as update_command
from .view import cases as cases_command, index as index_command, variants as variants_command
from .wipe import wipe as wipe_command



import logging

import click
import coloredlogs
import yaml
from loqusdb import __version__
from loqusdb.constants import GRCH37, GRCH38
from mongo_adapter import get_client
from mongo_adapter.exceptions import Error as DB_Error
from pymongo import uri_parser

from loqusdb.plugins.mongo.adapter import MongoAdapter

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

LOG = logging.getLogger(__name__)


@click.group()
@click.option(
    "-db",
    "--database",
    help="Defaults to 'loqusdb' if not specified",
)
@click.option("-u", "--username", type=str)
@click.option("-p", "--password", type=str)
@click.option(
    "-a",
    "--authdb",
    type=str,
    help="If authentication should be done against another database than --database",
)
@click.option(
    "-port",
    "--port",
    default=27017,
    show_default=True,
    help="Specify the port where to look for the mongo database.",
)
@click.option(
    "-h",
    "--host",
    default="localhost",
    show_default=True,
    help="Specify the host where to look for the mongo database.",
)
@click.option("--uri", help="Specify a mongodb uri")
@click.option("-c", "--config", type=click.File("r"), help="Use a config with db information")
@click.option(
    "-t", "--test", is_flag=True, help="Used for testing. This will use a mongomock database."
)
@click.option(
    "-g",
    "--genome-build",
    default="GRCh37",
    show_default=True,
    type=click.Choice([GRCH37, GRCH38]),
    help="Specify what genome build to use",
)
@click.option(
    "--keep-chr-prefix",
    is_flag=True,
    default=False,
    show_default=True,
    help="Retain the 'chr/Chr/CHR' prefix for chromosomes if it is present",
)
@click.option("-v", "--verbose", is_flag=True)
@click.version_option(__version__)
@click.pass_context
def cli(
    ctx,
    database,
    username,
    password,
    authdb,
    port,
    host,
    uri,
    verbose,
    config,
    test,
    genome_build,
    keep_chr_prefix,
):
    """loqusdb: manage a local variant count database."""
    loglevel = "INFO"
    if verbose:
        loglevel = "DEBUG"
    coloredlogs.install(level=loglevel)
    LOG.info("Running loqusdb version %s", __version__)

    configs = {}
    if config:
        try:
            configs = yaml.safe_load(config)
        except yaml.YAMLError as err:
            LOG.warning(err)
            ctx.abort()

    uri = configs.get("uri") or uri
    if test:
        uri = "mongomock://"
    try:
        client = get_client(
            host=configs.get("host") or host,
            port=configs.get("port") or port,
            username=configs.get("username") or username,
            password=configs.get("password") or password,
            authdb=authdb or database or "loqusdb",
            uri=uri,
        )
    except DB_Error as err:
        LOG.warning(err)
        ctx.abort()

    database = configs.get("db_name") or database

    if not database:
        database = "loqusdb"
        if uri:
            uri_info = uri_parser.parse_uri(uri)
            database = uri_info.get("database")

    adapter = MongoAdapter(client, db_name=database)

    genome_build = genome_build or configs.get("genome_build")
    keep_chr_prefix = keep_chr_prefix or configs.get("keep_chr_prefix")

    ctx.obj = {}
    ctx.obj["db"] = database
    if uri:
        ctx.obj["uri"] = uri
    else:
        ctx.obj["port"] = port
        ctx.obj["host"] = host
    ctx.obj["adapter"] = adapter
    ctx.obj["version"] = __version__
    ctx.obj["genome_build"] = genome_build
    ctx.obj["keep_chr_prefix"] = keep_chr_prefix



import logging
from datetime import datetime

import click
from loqusdb.constants import CHROMOSOMES, GRCH37, GRCH38
from loqusdb.utils.variant import format_variant
from vcftoolbox import HeaderParser, print_headers, print_variant

from loqusdb.commands.cli import cli as base_command

LOG = logging.getLogger(__name__)


@base_command.command("export", short_help="Export variants to VCF format")
@click.option(
    "-o",
    "--outfile",
    type=click.File("w"),
    help="Specify the path to a file where results should be stored.",
)
@click.option(
    "-t",
    "--variant-type",
    type=click.Choice(["sv", "snv"]),
    default="snv",
    show_default=True,
    help="If svs or snvs should be exported",
)
@click.option(
    "-f", "--freq", is_flag=True, help="Include observation frequencies in the exported VCF"
)
@click.pass_context
def export(ctx, outfile, variant_type, freq):
    """Export the variants of a loqus db

    The variants are exported to a vcf file
    """
    adapter = ctx.obj["adapter"]
    version = ctx.obj["version"]

    LOG.info("Export the variants from {0}".format(adapter))

    is_sv = variant_type == "sv"
    existing_chromosomes = set(adapter.get_chromosomes(sv=is_sv))

    genome = ctx.obj["genome_build"]
    chromosome_order = CHROMOSOMES[genome]
    keep_chr_prefix = ctx.obj["keep_chr_prefix"]

    ordered_chromosomes = []
    for chrom in chromosome_order:
        if keep_chr_prefix and chrom in existing_chromosomes:
            ordered_chromosomes.append(chrom)
            existing_chromosomes.remove(chrom)
        elif not keep_chr_prefix:
            if genome == GRCH37 and chrom in existing_chromosomes:
                ordered_chromosomes.append(chrom)
                existing_chromosomes.remove(chrom)
            elif genome == GRCH38 and chrom[3:] in existing_chromosomes:
                ordered_chromosomes.append(chrom)
                existing_chromosomes.remove(chrom)
    for chrom in existing_chromosomes:
        ordered_chromosomes.append(chrom)

    if variant_type == "snv":
        nr_cases = adapter.nr_cases(snv_cases=True)
    elif variant_type == "sv":
        nr_cases = adapter.nr_cases(sv_cases=True)
    else:
        raise ValueError(f"Unknown variant_type: {variant_type}, expected 'snv' or 'sv'")

    LOG.info(f"Found {nr_cases} cases in database")

    head = HeaderParser()
    head.add_fileformat("VCFv4.3")
    head.add_meta_line("NrCases", nr_cases)
    if freq:
        head.add_info(
            "Frq",
            "1",
            "Float",
            f"Observation frequency of the variant (not allele frequency) based on {nr_cases} cases",
        )
    head.add_info("Obs", "1", "Integer", "The number of observations for the variant")
    head.add_info("Hom", "1", "Integer", "The number of observed homozygotes")
    head.add_info("Hem", "1", "Integer", "The number of observed hemizygotes")
    head.add_version_tracking("loqusdb", version, datetime.now().strftime("%Y-%m-%d %H:%M"))

    if variant_type == "sv":
        head.add_info("END", "1", "Integer", "End position of the variant")
        head.add_info("SVTYPE", "1", "String", "Type of structural variant")
        head.add_info("SVLEN", "1", "Integer", "Length of structural variant")

    for chrom in ordered_chromosomes:
        length = adapter.get_max_position(chrom)
        head.add_contig(contig_id=chrom, length=str(length))

    print_headers(head, outfile=outfile)

    for chrom in ordered_chromosomes:
        if variant_type == "snv":
            LOG.info("Collecting all SNV variants")
            variants = adapter.get_variants(chromosome=chrom)
        else:
            LOG.info("Collecting all SV variants")
            variants = adapter.get_sv_variants(chromosome=chrom)
        LOG.info(f"{adapter.nr_variants(chromosome=chrom)} variants found")
        for variant in variants:
            variant_line = format_variant(
                variant, variant_type=variant_type, nr_cases=nr_cases, add_freq=freq
            )
            print_variant(variant_line=variant_line, outfile=outfile)
